# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import json
from pathlib import Path
from typing import Optional

from mozphab.conduit import (
    conduit,
)
from mozphab.exceptions import (
    Error,
)
from mozphab.logger import (
    logger,
)
from mozphab.repository import (
    Repository,
)

from .submit import (
    add_submit_arguments,
    submit,
)

UPLIFT_REPOSITORY_TAG_SLUG = "uplift"


def map_train_arg_to_repo(train_repo_shortname: str) -> dict:
    """Attempt to map the value of `--train` to a Phabricator repo."""
    repo = conduit.get_repository_by_shortname(train_repo_shortname)
    if not repo:
        raise Error(f"No repo named {train_repo_shortname} on Phabricator!")

    uplift_repos = conduit.get_repositories_with_tag(UPLIFT_REPOSITORY_TAG_SLUG)
    if not uplift_repos:
        raise Error("No uplift repos found on Phabricator! Please file a bug.")

    uplift_repo_shortnames = {uplift["fields"]["shortName"] for uplift in uplift_repos}

    if repo["fields"]["shortName"] not in uplift_repo_shortnames:
        raise Error(f"Repo {train_repo_shortname} is not an uplift repository.")

    return repo


def list_trains():
    """List all uplift repositories on Phabricator.

    Queries the Phabricator repository search endpoint for all repositories
    with the `uplift` tag set on them, and displays their shortnames as a list
    of valid values for `--train`.
    """
    repositories = conduit.get_repositories_with_tag(UPLIFT_REPOSITORY_TAG_SLUG)
    if not repositories:
        raise Error("Found no repos with `uplift` tag on Phabricator.")

    logger.info("Available trains for uplift:")
    for repository in repositories:
        logger.info(f"   - {repository['fields']['shortName']}")


def build_assessment_linking_url(
    lando_url: str, tip_commit_id: int, assessment_id: Optional[int] = None
) -> str:
    """Return the URL for linking revisions."""
    # Ensure `lando_url` doesn't have a trailing slash.
    base_url = lando_url.rstrip("/")

    # Build the URL with `revisions` parameter.
    url = f"{base_url}/uplift/request/?revisions={tip_commit_id}"

    # Add `assessment_id` if provided.
    if assessment_id is not None:
        url += f"&assessment_id={assessment_id}"

    return url


def uplift(repo: Repository, args: argparse.Namespace):
    if args.list_trains:
        return list_trains()

    if not args.train:
        raise Error("Missing `--train` argument!")

    phab_repo = map_train_arg_to_repo(train_repo_shortname=args.train)

    # When using `moz-phab uplift`, we perform a standard `submit` but with all
    # the in-repo configurations for submit destination overwritten to submit
    # to the uplift target repository instead.
    repo.call_sign = phab_repo["fields"]["callsign"]
    repo._phab_repo = phab_repo

    # Run the usual submit command with our patched arg values.
    commits = submit(repo, args)

    if commits:
        tip_commit = commits[-1]
        tip_commit_id = tip_commit.rev_id

        uplift_assessment_linking_url = build_assessment_linking_url(
            repo.lando_url, tip_commit_id, args.assessment_id
        )

        logger.warning(
            f"\nPlease navigate to {uplift_assessment_linking_url} and complete the uplift "
            "request form."
        )

    # Output machine-readable data to a file.
    if args.output_file:
        data = {"commits": [commit.to_dict() for commit in commits]}

        with args.output_file.open("w") as f:
            json.dump(data, f)


def add_parser(parser):
    uplift_parser = parser.add_parser(
        "uplift",
        help="Submit uplift request commit(s) to Phabricator.",
        description=(
            "MozPhab will create a new revision to request patches be uplifted "
            "to stable release trains.\n"
            "\n"
            "See https://wiki.mozilla.org/index.php?title="
            "Release_Management/Requesting_an_Uplift\n"
            "for more information."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Add all the same arguments as `moz-phab submit`.
    add_submit_arguments(uplift_parser)

    # Add uplift-specific arguments.
    uplift_parser.add_argument(
        "--train",
        help=(
            "Indicate Phabricator shortname of the release train this stack "
            "should be uplifted to (see `--list-trains`)."
        ),
    )
    uplift_parser.add_argument(
        "--list-trains",
        help="List all possible values for `--train`.",
        action="store_true",
    )
    uplift_parser.add_argument(
        "--no-rebase",
        help=(
            "Send the specified range of commits as-is for uplift - do not attempt "
            "to rebase."
        ),
        action="store_true",
    )
    uplift_parser.add_argument(
        "--assessment-id",
        help="Existing assessment ID to link the new uplift revision to.",
    )
    uplift_parser.add_argument(
        "--output-file",
        dest="output_file",
        type=Path,
        help=argparse.SUPPRESS,
    )
    uplift_parser.set_defaults(func=uplift, needs_repo=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import re
from dataclasses import (
    asdict,
    dataclass,
    field,
)
from typing import (
    Dict,
    List,
    Optional,
)

from mozphab.logger import logger

ARC_COMMIT_DESC_TEMPLATE = """
{title}

Summary:
{body}

Test Plan:

Reviewers: {reviewers}

Subscribers:

Bug #: {bug_id}
""".strip()

WIP_RE = re.compile(r"^(?:WIP[: ]|WIP$)", flags=re.IGNORECASE)


@dataclass
class Commit:
    """`moz-phab`'s representation of a commit."""

    name: str = ""
    node: str = ""
    orig_node: str = ""
    submit: bool = False
    title: str = ""
    title_preview: str = ""
    body: str = ""
    author_date_epoch: int = 0
    author_name: str = ""
    author_email: str = ""

    author_date: Optional[str] = None
    parent: Optional[str] = None
    bug_id: Optional[str] = None
    bug_id_orig: Optional[str] = None
    rev_id: Optional[int] = None
    rev_phid: Optional[str] = None
    wip: Optional[bool] = None
    tree_hash: Optional[str] = None
    reviewers: Dict[str, List[str]] = field(default_factory=dict)

    @property
    def has_reviewers(self) -> bool:
        """Return `True` if the commit has reviewers."""
        if not self.reviewers:
            return False

        return bool(self.reviewers.get("granted") or self.reviewers.get("request"))

    def build_arc_commit_message(self) -> str:
        """Contruct a commit message for this `Commit` object."""
        # WIP submissions shouldn't set reviewers.
        if self.wip:
            reviewers = ""
        else:
            reviewers = ", ".join(self.reviewers["granted"] + self.reviewers["request"])

        # Create arc-annotated commit description.
        template_vars = {
            "title": self.revision_title(),
            "body": self.body,
            "reviewers": reviewers,
            # Change `None` to an empty string.
            "bug_id": self.bug_id or "",
        }

        message = ARC_COMMIT_DESC_TEMPLATE.format(**template_vars)
        logger.debug("--- arc message\n%s\n---" % message)
        return message

    def revision_title(self) -> str:
        """Returns a string suitable for a Revision title for the given commit."""
        title = WIP_RE.sub("", self.title_preview).lstrip()
        if self.wip:
            title = f"WIP: {title}"
        return title

    def wip_in_commit_title(self) -> bool:
        """Return `True` if the commit title indicates the revision is a WIP."""
        return WIP_RE.search(self.title) is not None

    @property
    def message(self) -> str:
        return f"{self.title}\n\n{self.body}"

    def to_dict(self) -> dict:
        """Convert the `Commit` to a `dict`."""
        return asdict(self)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
from typing import Optional
from unittest import mock

import pytest

from mozphab import exceptions, helpers, mozphab
from mozphab.commands import patch
from mozphab.config import Config


def test_resolve_branch_name():
    class Args:
        def __init__(self, name: Optional[str] = None, no_commit: bool = False):
            self.name = name
            self.no_commit = no_commit

    rev_id = "123"

    args = Args(name="branch")
    config = Config(should_access_file=False)

    assert (
        patch.resolve_branch_name(args, config, rev_id) == "branch"
    ), "Branch name passed via args should take precedent."

    args = Args(no_commit=True)
    assert (
        patch.resolve_branch_name(args, config, rev_id) is None
    ), "No branch name should be used when `no_commit` is `True`."

    args = Args()
    assert (
        patch.resolve_branch_name(args, config, rev_id) == "phab-D123"
    ), "Branch name should match the default when not changed."

    config.branch_name_template = "phab/D{rev_id}"
    assert (
        patch.resolve_branch_name(args, config, rev_id) == "phab/D123"
    ), "Branch name should change based on the configuration."

    config.branch_name_template = "phabrev"
    assert (
        patch.resolve_branch_name(args, config, rev_id) == "phabrev"
    ), "Using template strings should be optional."

    config.create_commit = False
    assert (
        patch.resolve_branch_name(args, config, rev_id) is None
    ), "When create_commit is False in config no branch should be used."
    config.create_commit = True


def test_check_revision_id():
    check_revision_id = patch.check_revision_id

    assert check_revision_id("123") == 123
    assert check_revision_id("D123") == 123
    assert check_revision_id("https://phabricator.example.com/D123") == 123
    assert check_revision_id("https://phabricator.example.com/D123?") == 123
    with pytest.raises(argparse.ArgumentTypeError):
        check_revision_id("D")
    with pytest.raises(argparse.ArgumentTypeError):
        check_revision_id("https://example.com/")


@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
def test_get_diff_by_id(m_get_diffs):
    args = {}
    with pytest.raises(exceptions.Error):
        patch.get_diff_by_id(args.get("diff_id"))

    args = {"diff_id": 1}
    m_get_diffs.return_value = {"DIFFPHID-1": {"id": 1}}
    phid, diff = patch.get_diff_by_id(args.get("diff_id"))
    assert phid == "DIFFPHID-1", "Should return PHID as first element"
    assert diff["id"] == 1, "Should return diff dict as second element"

    m_get_diffs.return_value = {}
    with pytest.raises(exceptions.NotFoundError):
        patch.get_diff_by_id(args.get("diff_id"))


def test_update_revision_with_new_diff():
    revs = [REV_1]
    patch.update_revision_with_new_diff(revs, DIFF_3)
    assert (
        revs[0]["fields"]["diffPHID"] == "DIFFPHID-3"
    ), "Should update related revision"

    revs = [REV_1, REV_2]
    patch.update_revision_with_new_diff(revs, DIFF_1)
    assert (
        revs[0]["fields"]["diffPHID"] == "DIFFPHID-1"
    ), "Should update related revision when given multiple revisions"

    # Should raise an error for unrelated diff
    with pytest.raises(exceptions.Error):
        patch.update_revision_with_new_diff(revs, DIFF_4)


def test_strip_depends_on():
    strip = helpers.strip_depends_on

    assert "" == strip("Depends on D123")
    assert "" == strip("\n Depends on D1\n")
    assert "title" == strip("title\n\nDepends on D1")
    assert "title\n\nbefore\n\nafter" == strip("title\n\nbefore\nDepends on D1\nafter")
    assert "Depends on DA" == strip("Depends on DA")


def test_prepare_body():
    prep = helpers.prepare_body
    assert "\n\nDifferential Revision: http://phabricator.test/D1" == prep(
        "", "", 1, "http://phabricator.test"
    )
    assert "\n\nDifferential Revision: http://phabricator.test/D2" == prep(
        "", "", 2, "http://phabricator.test"
    )
    assert "title\n\n" "Differential Revision: http://phabricator.test/D2" == prep(
        "title", "", 2, "http://phabricator.test"
    )
    assert (
        "title\n\n"
        "some\n"
        "summary\n\n"
        "Differential Revision: http://phabricator.test/D2"
        == prep("title", "some\nsummary", 2, "http://phabricator.test")
    )


@mock.patch("mozphab.gitcommand.which_path")
@mock.patch("mozphab.gitcommand.check_call")
def test_apply_patch(m_check_call, _):
    patch.config.git_command = ["git"]
    patch.apply_patch("diff", "x")
    m_check_call.assert_called_once()


def test_base_ref():
    assert patch.get_base_ref({"fields": {}}) is None

    diff = {"fields": {"refs": [{"identifier": "sha1", "type": "sometype"}]}}
    assert patch.get_base_ref(diff) is None

    diff = {"fields": {"refs": [{"identifier": "sha1", "type": "base"}]}}
    assert patch.get_base_ref(diff) == "sha1"


@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.git.Git.is_worktree_clean")
@mock.patch("mozphab.commands.patch.config")
@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.get_ancestor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_successor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
@mock.patch("mozphab.commands.patch.get_base_ref")
@mock.patch("mozphab.git.Git.before_patch")
@mock.patch("mozphab.commands.patch.apply_patch")
@mock.patch("mozphab.commands.patch.prepare_body")
@mock.patch("mozphab.git.Git.apply_patch")
@mock.patch("mozphab.git.Git.check_node")
@mock.patch("builtins.print")
def test_patch(
    m_print,
    m_git_check_node,
    m_git_apply_patch,
    m_prepare_body,
    m_apply_patch,
    m_git_before_patch,
    m_get_base_ref,
    m_get_diffs,
    m_get_successor_phids,
    m_get_ancestor_phids,
    m_get_revisions,
    m_git_check_conduit,
    m_config,
    m_git_is_worktree_clean,
    m_call_conduit,
    git,
):
    mozphab.conduit.set_repo(git)

    m_git_check_conduit.return_value = False
    m_config.arc_command = "arc"
    m_config.branch_name_template = "phab-D{rev_id}"
    with pytest.raises(exceptions.Error):
        patch.patch(git, None)

    class Args:
        def __init__(
            self,
            revision_id=123,
            no_commit=False,
            raw=False,
            apply_to="base",
            yes=False,
            skip_dependencies=False,
            include_abandoned=False,
            force_vcs=False,
            name=None,
            diff_id=None,
        ):
            self.revision_id = revision_id
            self.no_commit = no_commit
            self.raw = raw
            self.apply_to = apply_to
            self.yes = yes
            self.skip_dependencies = skip_dependencies
            self.include_abandoned = include_abandoned
            self.force_vcs = force_vcs
            self.name = name
            self.diff_id = diff_id

    git.args = Args()
    m_git_check_conduit.return_value = True
    m_git_is_worktree_clean.return_value = False
    with pytest.raises(exceptions.Error):
        patch.patch(git, git.args)

    m_git_is_worktree_clean.return_value = True
    m_get_revisions.return_value = []
    with pytest.raises(exceptions.Error):
        patch.patch(git, git.args)

    m_config.always_full_stack = False
    m_get_successor_phids.return_value = []
    m_get_ancestor_phids.return_value = []
    m_get_base_ref.return_value = "sha111"
    m_call_conduit.return_value = "raw"  # differential.getrawdiff
    m_get_revisions.return_value = [
        {
            "phid": "PHID-1",
            "id": 1,
            "fields": {
                "diffPHID": "DIFFPHID-1",
                "title": "title",
                "summary": "summary",
            },
        }
    ]
    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1}
    m_git_check_node.return_value = "sha111"
    m_prepare_body.return_value = "commit message"
    patch.patch(git, git.args)
    m_git_apply_patch.assert_called_once_with(
        "raw",
        "commit message",
        "user <author@example.com>",
        1547806078,
    )
    m_apply_patch.assert_not_called()
    m_get_base_ref.assert_called_once()
    m_git_before_patch.assert_called_once_with("sha111", "phab-D1")

    m_git_apply_patch.reset_mock()
    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1}
    patch.patch(git, git.args)
    m_git_apply_patch.assert_called_once_with(
        "raw",
        "commit message",
        "user <author@example.com>",
        1547806078,
    )

    # --diff-id
    m_get_diffs.side_effect = [
        {"DIFFPHID-1": DIFF_1},
        {"DIFFPHID-3": DIFF_3},
    ]
    git.args = Args(diff_id=3)
    patch.patch(git, git.args)
    m_git_apply_patch.assert_called_with(
        "raw",
        "commit message",
        "user 3 <author@example.com>",
        1547806078,
    )
    m_get_diffs.side_effect = None

    # --diff-id raises NotFoundError
    m_get_diffs.return_value = {}
    git.args = Args(diff_id=100)
    with pytest.raises(exceptions.NotFoundError):
        patch.patch(git, git.args)

    # --diff-id raises Error when belonging to different rev
    m_get_diffs.side_effect = [
        {"DIFFPHID-1": DIFF_1},
        {"DIFFPHID-4": DIFF_4},
    ]
    git.args = Args(diff_id=4)
    with pytest.raises(exceptions.Error):
        patch.patch(git, git.args)
    m_get_diffs.side_effect = None

    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1}
    m_get_base_ref.return_value = None
    with pytest.raises(exceptions.Error):
        patch.patch(git, git.args)

    m_get_base_ref.return_value = "sha111"
    m_git_apply_patch.reset_mock()
    m_git_before_patch.reset_mock()
    # --raw
    git.args = Args(raw=True)
    patch.patch(git, git.args)
    m_git_before_patch.assert_not_called()
    m_git_apply_patch.assert_not_called()
    m_apply_patch.assert_not_called()
    m_print.assert_called_with("raw")

    # skip-dependencies
    git.args = Args(raw=True, skip_dependencies=True)
    m_get_successor_phids.reset_mock()
    patch.patch(git, git.args)
    m_get_successor_phids.assert_not_called()

    m_git_before_patch.reset_mock()
    # --no_commit
    git.args = Args(no_commit=True)
    patch.patch(git, git.args)
    m_git_before_patch.assert_called_once()
    m_git_apply_patch.assert_not_called()
    m_apply_patch.assert_called_once_with("raw", "x")
    m_git_before_patch.assert_called_once_with("sha111", None)

    m_apply_patch.reset_mock()
    m_git_before_patch.reset_mock()
    # create_commit=False in config
    m_config.create_commit = False
    git.args = Args()
    patch.patch(git, git.args)
    m_git_before_patch.assert_called_once()
    m_git_apply_patch.assert_not_called()
    m_apply_patch.assert_called_once_with("raw", "x")
    m_git_before_patch.assert_called_once_with("sha111", None)

    m_apply_patch.reset_mock()
    m_config.create_commit = True
    m_git_before_patch.reset_mock()
    # --no_commit --applyto head
    git.args = Args(no_commit=True, apply_to="head")
    patch.patch(git, git.args)

    m_get_base_ref.reset_mock()
    m_apply_patch.reset_mock()
    # --apply_to head
    git.args = Args(apply_to="head")
    patch.patch(git, git.args)
    m_get_base_ref.assert_not_called()
    m_git_apply_patch.assert_called_once_with(
        "raw",
        "commit message",
        "user <author@example.com>",
        1547806078,
    )
    m_apply_patch.assert_not_called()

    m_git_before_patch.reset_mock()
    node = "abcdef"
    m_git_check_node.return_value = node
    # --applyto NODE
    git.args = Args(apply_to=node)
    patch.patch(git, git.args)
    m_git_before_patch.assert_called_once_with(node, "phab-D1")

    m_git_before_patch.reset_mock()
    # --applyto here
    git.args = Args(apply_to="here")
    patch.patch(git, git.args)
    m_git_before_patch.assert_called_once_with(None, "phab-D1")

    m_git_before_patch.reset_mock()
    # --name NAME
    git.args = Args(name="feature")
    patch.patch(git, git.args)
    m_git_before_patch.assert_called_once_with(node, "feature")

    # ########## no commit info in diffs
    m_get_diffs.return_value = {
        "DIFFPHID-1": {"id": 1, "attachments": {"commits": {"commits": []}}}
    }
    m_call_conduit.side_effect = ("raw",)
    git.args = Args()
    with pytest.raises(exceptions.Error):
        patch.patch(git, git.args)

    m_git_apply_patch.reset_mock()
    m_apply_patch.reset_mock()
    git.args = Args(no_commit=True)
    patch.patch(git, git.args)
    m_git_apply_patch.assert_not_called()
    m_apply_patch.assert_called_once()

    m_print.reset_mock()
    m_call_conduit.side_effect = ("raw",)
    git.args = Args(raw=True)
    patch.patch(git, git.args)
    m_print.assert_called_once_with("raw")

    # ########## multiple revisions
    m_print.reset_mock()
    m_get_revisions.side_effect = ([REV_1], [REV_2])
    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1, "DIFFPHID-2": DIFF_2}
    m_get_ancestor_phids.return_value = ["PHID-2"]
    m_call_conduit.side_effect = ("raw2", "raw1")
    # --raw 2 revisions in stack
    patch.patch(git, git.args)
    m_print.assert_has_calls((mock.call("raw2"), mock.call("raw1")))

    # node not found
    m_get_revisions.side_effect = None
    m_git_check_node.side_effect = exceptions.NotFoundError
    git.args = Args(apply_to=node)
    with pytest.raises(exceptions.Error) as e:
        patch.patch(git, git.args)
        assert "Unknown revision: %s\nERROR" % node in e.msg

    # successors
    m_get_revisions.reset_mock()
    m_get_revisions.side_effect = ([REV_1], [REV_2])
    m_get_successor_phids.side_effect = (["PHID-2"], ["PHID-2"], [])
    m_get_ancestor_phids.return_value = []
    m_call_conduit.side_effect = ("raw2", "raw1")
    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1, "DIFFPHID-2": DIFF_2}
    git.args = Args(revision_id=1, raw=True, yes=True)
    patch.patch(git, git.args)
    assert m_get_revisions.call_args_list == [
        mock.call(ids=[1]),
        mock.call(phids=["PHID-2"]),
    ]

    # multiple successors
    m_get_revisions.reset_mock()
    m_get_revisions.side_effect = ([REV_1], [REV_2])
    m_get_successor_phids.side_effect = (exceptions.NonLinearException,)
    m_call_conduit.side_effect = ("raw",)
    m_get_diffs.return_value = {"DIFFPHID-1": DIFF_1}
    patch.patch(git, git.args)
    m_get_revisions.assert_called_once_with(ids=[1])


REV_1 = {
    "phid": "PHID-1",
    "id": 1,
    "fields": {"diffPHID": "DIFFPHID-1", "title": "title", "summary": "summary"},
}

REV_2 = {
    "phid": "PHID-2",
    "id": 2,
    "fields": {"diffPHID": "DIFFPHID-2", "title": "title", "summary": "summary"},
}

DIFF_1 = {
    "id": 1,
    "phid": "DIFFPHID-1",
    "fields": {"revisionPHID": "PHID-1", "dateCreated": 1547806078},
    "attachments": {
        "commits": {
            "commits": [
                {
                    "author": {
                        "name": "user",
                        "email": "author@example.com",
                        "epoch": 1547806078,
                    }
                }
            ]
        }
    },
}

DIFF_2 = {
    "id": 2,
    "phid": "DIFFPHID-2",
    "attachments": {
        "commits": {
            "commits": [{"author": {"name": "user", "email": "author@example.com"}}]
        }
    },
}

DIFF_3 = {
    "id": 3,
    "phid": "DIFFPHID-3",
    "fields": {"revisionPHID": "PHID-1", "dateCreated": 1547806078},
    "attachments": {
        "commits": {
            "commits": [
                {
                    "author": {
                        "name": "user 3",
                        "email": "author@example.com",
                        "epoch": 1547806078,
                    }
                }
            ]
        }
    },
}

DIFF_4 = {
    "id": 4,
    "phid": "DIFFPHID-4",
    "fields": {"revisionPHID": "PHID-100"},
    "attachments": {
        "commits": {
            "commits": [{"author": {"name": "user", "email": "author@example.com"}}]
        }
    },
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest import mock

from mozphab.args import should_fallback_to_submit


@mock.patch("mozphab.args.find_repo_root")
def test_should_fallback_to_submit(m_repo_root):
    commands = {"submit", "install-certificate"}

    m_repo_root.return_value = "someroot"

    argv = []
    assert should_fallback_to_submit(
        argv, commands
    ), "Empty run of `moz-phab` should fallback."

    argv = ["-h"]
    assert not should_fallback_to_submit(
        argv, commands
    ), "Running `moz-phab` with `-h` should not fallback."

    argv = ["--help"]
    assert not should_fallback_to_submit(
        argv, commands
    ), "Running `moz-phab` with `--help` should not fallback."

    argv = ["blah", "-h"]
    assert not should_fallback_to_submit(
        argv, commands
    ), "Running `moz-phab` with `-h` should not fallback."

    argv = ["blah", "--help"]
    assert not should_fallback_to_submit(
        argv, commands
    ), "Running `moz-phab` with `--help` should not fallback."

    argv = ["bad-command", "argument"]
    assert should_fallback_to_submit(
        argv, commands
    ), "Running `moz-phab` with unknown command should result in fallback."

    m_repo_root.return_value = None
    assert not should_fallback_to_submit(argv, commands), (
        "Running `moz-phab` with unknown command outside of repo root should "
        "not result in fallback."
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import annotations

import argparse
import json
import mimetypes
import os
import re
import sys
import time
import uuid
from contextlib import suppress
from functools import lru_cache
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

import hglib
from packaging.version import Version

from mozphab import environment

from .commits import Commit
from .config import config
from .diff import Diff
from .exceptions import CommandError, Error, NotFoundError
from .helpers import (
    create_hunk_lines,
    is_valid_email,
    parse_config,
    short_node,
    temporary_file,
    which_path,
)
from .logger import logger
from .repository import Repository
from .spinner import clear_terminal_line, wait_message
from .subprocess_wrapper import debug_log_command
from .telemetry import telemetry

MINIMUM_MERCURIAL_VERSION = Version("4.3.3")


class Mercurial(Repository):
    def __init__(self, path: str):
        dot_path = os.path.join(path, ".hg")
        if not os.path.isdir(dot_path):
            raise ValueError("%s: not a hg repository" % path)
        logger.debug("found hg repo in %s", path)

        super().__init__(path, dot_path)
        self.vcs = "hg"
        self._hg_binary = config.hg_command[0]
        self.revset = None
        self.strip_nodes = []
        self.status = None
        self.obsstore = None
        self.unlink_obsstore = False
        self.use_evolve = False
        self.use_topic = False
        self.has_mq = False
        self.has_shelve = False
        self.previous_bookmark = None
        self.has_temporary_bookmark = False
        self.username = ""

        # Check for `hg` presence
        if not which_path(self._hg_binary):
            raise Error(f"Failed to find hg executable ({self._hg_binary})")

        self._config_options = {}
        self._safe_config_options = {}
        self._extra_options = {}
        self._safe_mode = False
        self._repo_path = path
        hglib.HGPATH = self._hg_binary
        self._repo = None
        self._configs = []
        major, minor, micro, *_ = self.repository.version
        self.mercurial_version = Version(f"{major}.{minor}.{micro}")
        self.vcs_version = str(self.mercurial_version)
        if self.mercurial_version < MINIMUM_MERCURIAL_VERSION:
            raise Error(
                f"You are currently running Mercurial {self.mercurial_version}.  "
                f"Mercurial {MINIMUM_MERCURIAL_VERSION} or newer is required."
            )

        # Some Mercurial commands require the current working directory to be
        # set to the repository root (eg. `hg files`).
        os.chdir(self._repo_path)

    @property
    def repository(self) -> hglib.hgclient:
        """Returns the hglib.hgclient instance.

        If the config has changed, recreate the instance.
        """
        configs = [f"{key}={value}" for key, value in self._get_config_options()]
        configs.sort()
        # returns the repo instance if the config has not changed.
        if self._repo is not None and self._configs == configs:
            return self._repo

        if self._repo:
            self._repo.close()

        self._configs = configs
        self._repo = hglib.open(self._repo_path, encoding="UTF-8", configs=configs)
        return self._repo

    def _get_config_options(self) -> List[tuple]:
        """Returns the --config options for hg

        Updated with the safe ones when safe mode is used.
        """
        options = dict(self._config_options)
        if self._safe_mode:
            options.update(self._safe_config_options)
        return list(options.items())

    @classmethod
    def is_repo(cls, path: str) -> bool:
        """Quick check for repository at specified path."""
        return os.path.exists(os.path.join(path, ".hg"))

    @staticmethod
    def _get_extension(extension: str, hg_config: dict) -> Optional[str]:
        for prefix in ("extensions.%s", "extensions.hgext.%s"):
            field = prefix % extension
            if field in hg_config:
                return hg_config.get(field, "")
        return None

    @staticmethod
    def _get_extensions(
        *,
        from_config: Optional[List[str]] = None,
        from_args: Optional[List[str]] = None,
    ) -> List[str]:
        assert from_config or from_args

        extensions = []
        if from_config:
            for name in from_config:
                if name.startswith("extensions."):
                    extensions.append(re.sub(r"^extensions\.(?:hgext\.)?", "", name))

        else:
            args = from_args.copy()
            while len(args) >= 2:
                arg = args.pop(0)
                if arg != "--config":
                    continue
                arg = args.pop(0)
                if arg.startswith("extensions."):
                    name, value = arg.split("=", maxsplit=1)
                    extensions.append(re.sub(r"^extensions\.(?:hgext\.)?", "", name))

        return sorted(extensions)

    def is_worktree_clean(self) -> bool:
        status = self._status()
        return not status["T"]

    def hg(self, command: List[str], **kwargs):
        self.hg_out(command, capture=False, **kwargs)

    def hg_out(
        self,
        command: List[str],
        capture: bool = True,
        expect_binary: bool = False,
        strip: bool = True,
        split: bool = True,
        never_log: bool = False,
    ) -> Optional[bytes | str]:
        def error_handler(exit_code, stdout, stderr):
            if not capture:
                clear_terminal_line()
                if stderr:
                    print(stderr.decode(), file=sys.stderr, end="")
                if stdout:
                    print(stdout.decode(), end="")

            if not never_log:
                if stderr:
                    logger.debug(stderr.decode().rstrip())
                if stdout:
                    logger.debug(stdout.decode().rstrip())

            raise CommandError(
                "command '%s' failed to complete successfully"
                % command_bytes[0].decode(),
                exit_code,
            )

        for arg, value in self._extra_options.items():
            command.extend([arg, value])

        debug_log_command(["hg"] + command)
        command_bytes = [c.encode() for c in command]
        out = self.repository.rawcommand(command_bytes, eh=error_handler)

        if expect_binary:
            logger.debug("%s bytes of data received", len(out))
            return out

        out = out.decode()
        if strip:
            out = out.rstrip()
        if out and not never_log:
            logger.debug(out)

        if capture:
            return out.splitlines() if split else out

        clear_terminal_line()
        print(out, end="")
        return None

    def hg_log(
        self, revset: str, split: bool = True, select: str = "node"
    ) -> Optional[str]:
        return self.hg_out(["log", "-T", "{%s}\n" % select, "-r", revset], split=split)

    def before_submit(self):
        self.validate_email()

        # Remember the currently checked out commit.  If a bookmark is active
        # just use that, otherwise create a randomly named bookmark which will
        # be deleted in cleanup(). Mercurial will automatically move the
        # bookmark to the successors as we update commits.
        for active, bookmark in [
            line.split(" ", 1)
            for line in self.hg_out(["bookmark", "-T", "{active} {bookmark}\n"])
        ]:
            if active == "True":
                self.previous_bookmark = bookmark
                break
        else:
            self.previous_bookmark = "moz-phab_%s" % self.hg_out(
                ["id", "-q"], split=False, strip=True
            )
            self.has_temporary_bookmark = True
            self.hg(["bookmark", self.previous_bookmark])

    def after_submit(self):
        # Restore the previously active commit.
        self.hg(["update", self.previous_bookmark, "--quiet"])

    def cleanup(self):
        # Remove the store of obsolescence markers; if the user doesn't have evolve
        # installed mercurial will warn if this exists.
        if not self.use_evolve and self.unlink_obsstore:
            with suppress(FileNotFoundError):
                os.unlink(self.obsstore)

        if self.strip_nodes:
            # With the obsstore deleted the amended nodes are no longer hidden, so
            # we need to strip them completely from the repo.
            self.hg(["strip", "--hidden"] + self.strip_nodes)
            self.strip_nodes = []

        # Remove temporary bookmark
        if self.has_temporary_bookmark:
            bookmarks = self.hg_out(["bookmark", "-T", "{bookmark}\n"])
            if self.previous_bookmark in bookmarks:
                self.hg(["bookmark", "--delete", self.previous_bookmark])
            self.previous_bookmark = None
            self.has_temporary_bookmark = False

    def _status(self) -> Dict[str, List[str]]:
        # `hg status` is slow on large repos.  As we'll need both uncommitted changes
        # and untracked files separately, run it once and cache results.
        if self.status is None:
            self.status = {"T": [], "U": []}
            for line in self.hg_out(
                ["status", "--added", "--deleted", "--modified", "--unknown"],
                split=True,
            ):
                status, path = line.split(" ", 1)
                self.status["U" if status == "?" else "T"].append(path)
        return self.status

    def untracked(self) -> List[str]:
        return self._status()["U"]

    def _refresh_commit(self, commit: Commit, node: str, rev: Optional[str] = None):
        """Update commit's node and name from node and rev."""
        if not rev:
            rev = self.hg_log(node, select="rev", split=False)
        commit.node = node
        commit.name = f"{rev}:{short_node(node)}"

    def _get_successor(self, node: str) -> Tuple[Optional[str], Optional[str]]:
        """Get the successor of the commit represented by its node.

        Returns: a tuple containing rev and node.
        """
        hg_log = self.hg_out(
            ["log"]
            + ["-T", "{rev} {node}\n"]
            + ["--hidden"]
            + ["-r", "successors(%s) and not obsolete()" % node]
        )
        if not hg_log:
            return None, None

        # Not sure the best way to handle multiple successors, so just bail out.
        if len(hg_log) > 1:
            raise Error("Multiple successors found for %s, unable to continue" % node)

        return hg_log[0].split(" ", 1)

    def refresh_commit_stack(self, commits: List[Commit]):
        """Update all commits to point to their superseded commit."""
        for commit in commits:
            (rev, node) = self._get_successor(commit.node)
            if rev and node:
                self._refresh_commit(commit, node, rev)

        self.revset = "%s::%s" % (commits[0].node, commits[-1].node)

    def set_args(self, args: argparse.Namespace):
        """Sets up the right environment for hg, prior to running it.

        Sets:
        - all the --config options (includes safe mode)
        - extra options
        - the `revset` attribute
        """
        super().set_args(args)

        # Load hg config into hg_config.  We'll specify specific settings on
        # the command line when calling hg; all other user settings are ignored.
        # Do not parse shell alias extensions.
        hg_config = parse_config(
            self.hg_out(["config"], never_log=True),
            lambda name, value: not (
                name.startswith("extensions.") and value.startswith("!")
            ),
        )

        self._safe_mode = self.args.safe_mode or config.safe_mode
        self._safe_config_options.clear()
        self._config_options.clear()
        self._extra_options.clear()

        # Need to use the correct username.
        if "ui.username" not in hg_config:
            raise Error("ui.username is not configured in your hgrc")
        self._safe_config_options["ui.username"] = hg_config["ui.username"]
        self.username = hg_config["ui.username"]

        # Always need rebase.
        self._config_options["extensions.rebase"] = ""

        # Mercurial should never paginate the response.
        self._extra_options["--pager"] = "never"

        # Perform rebases in-memory to improve performance (requires v4.5+).
        if not self._safe_mode and self.mercurial_version >= Version("4.5"):
            self._config_options["rebase.experimental.inmemory"] = "true"

        # Enable evolve if the user's currently using it.  evolve makes amending
        # commits with children trivial (amongst other things).
        ext_evolve = self._get_extension("evolve", hg_config)
        if ext_evolve is not None:
            self._safe_config_options["extensions.evolve"] = ext_evolve
            self.use_evolve = True

        # Otherwise just enable obsolescence markers, and when we're done remove
        # the obsstore we created.
        else:
            self._config_options["experimental.evolution.createmarkers"] = "true"
            self._config_options["extensions.strip"] = ""
            self.use_evolve = False
            self.obsstore = os.path.join(self.path, ".hg", "store", "obsstore")
            self.unlink_obsstore = not os.path.exists(self.obsstore)

            # Warn users that `evolve` is not enabled and direct
            # them to `mach vcs-setup`.
            logger.warning(
                "moz-phab detected that hg isn't configured with `evolve`. This "
                "usually causes severe performance issues. Run `./mach vcs-setup` "
                "to configure evolve."
            )

        # User may wish to use topics rather than bookmarks.
        ext_topic = self._get_extension("topic", hg_config)
        if ext_topic is not None:
            self._safe_config_options["extensions.topic"] = ext_topic
            self.use_topic = True

        # This script interacts poorly with mq.
        ext_mq = self._get_extension("mq", hg_config)
        self.has_mq = ext_mq is not None
        if self.has_mq:
            self._safe_config_options["extensions.mq"] = ext_mq

        # `shelve` is useful for dealing with uncommitted changes; track if it's
        # currently enabled so we can tailor our error accordingly.
        self.has_shelve = self._get_extension("shelve", hg_config) is not None

        # Disable the user's hgrc file, to ensure we run without rogue extensions.
        if self._safe_mode:
            os.environ["HGRCPATH"] = ""
            options = self._get_config_options()
            logger.debug(
                "hg extensions (safe mode): %s",
                ", ".join(self._get_extensions(from_args=options)),
            )
        else:
            logger.debug(
                "hg extensions: %s",
                ", ".join(self._get_extensions(from_config=hg_config)),
            )

        if hasattr(self.args, "start_rev"):
            is_single = hasattr(self.args, "single") and self.args.single
            # Set the default start revision.
            if self.args.start_rev == environment.DEFAULT_START_REV:
                if is_single:
                    start_rev = self.args.end_rev
                else:
                    start_rev = "ancestors(.) and not public() and not obsolete()"
            else:
                start_rev = self.args.start_rev

            # Resolve to nodes as that's nicer to read.
            try:
                start = self.hg_log(start_rev)[0]
            except IndexError:
                if self.args.start_rev == environment.DEFAULT_START_REV:
                    raise Error("Failed to find draft commits to submit")
                else:
                    raise Error(
                        "Failed to start of commit range: %s" % self.args.start_rev
                    )

            if is_single:
                self.revset = start
                return

            end_rev = self.args.end_rev

            try:
                end = self.hg_log(end_rev)[0]
            except IndexError:
                raise Error("Failed to end of commit range: %s" % end_rev)

            self.revset = "%s::%s" % (short_node(start), short_node(end))

    def commit_stack(self, **kwargs) -> List[Commit]:
        # Grab all the info we need about the commits, using randomness as a delimiter.
        boundary = "--%s--\n" % uuid.uuid4().hex
        hg_log = self.hg_out(
            ["log"]
            + [
                "-T",
                "{rev}\n{node}\n{date|hgdate}\n{author|person}\n{author|email}\n"
                "{desc}%s" % boundary,
            ]
            + ["-r", self.revset],
            split=False,
            strip=False,
        )[: -len(boundary)]

        # Guard against conditions where the stack is empty (see bug 1547083).
        if not hg_log:
            return []

        commits = []
        nodes = []
        branching_children = []
        for log_line in hg_log.split(boundary):
            rev, node, author_date, author_name, author_email, desc = log_line.split(
                "\n", 5
            )
            desc = desc.splitlines()

            children = self.hg_log("children(%s)" % node)
            if len(children) > 1 and not self.use_evolve:
                branching_children.extend(children)

            commits.append(
                Commit(
                    name=f"{rev}:{short_node(node)}",
                    node=node,
                    submit=True,
                    orig_node=node,
                    title=desc[0],
                    title_preview=desc[0],
                    body="\n".join(desc[1:]).rstrip(),
                    bug_id=None,
                    reviewers={"request": [], "granted": []},
                    rev_id=None,
                    author_date_epoch=int(author_date.split(" ")[0]),
                    author_name=author_name,
                    author_email=author_email,
                )
            )
            nodes.append(node)

        if branching_children:
            will_be_deleted = [
                short_node(c) for c in branching_children if c not in nodes
            ]
            msg = "following commits will be DELETED:\n%s" % will_be_deleted
            if not self.args.force_delete:
                raise Error(
                    "DAG branch point detected. Please install the evolve extension.\n"
                    "(https://www.mercurial-scm.org/doc/evolution/)\n"
                    "If you continue with `--force-delete` the %s" % msg
                )
            else:
                logger.warning("`--force-delete` used. The %s", msg)

        return commits

    def is_node(self, node: str) -> bool:
        try:
            self.hg_out(["identify", "-q", "-r", node])
        except CommandError:
            return False

        return True

    def check_node(self, node: str) -> str:
        if not self.is_node(node):
            raise NotFoundError()

        return node

    def checkout(self, node: str):
        self.hg(["update", "--quiet", node])

    def commit(self, body: str):
        """Commit the changes in the working directory."""
        with temporary_file(body) as temp_f:
            self.hg(["commit", "--logfile", temp_f])

    def before_patch(self, node: str, name: str):
        """Prepare repository to apply the patches.

        Args:
            node - SHA1 or revision of the base commit
            name - name of the bookmark/topic to be created
        """
        # Checkout sha
        if node:
            with wait_message("Checking out %s.." % short_node(node)):
                self.checkout(node)
            if not self.args.raw:
                logger.info("Checked out %s", short_node(node))

        if name and config.create_bookmark and not self.args.no_bookmark:
            bookmarks = self.hg_out(["bookmarks", "-T", "{bookmark}\n"])
            bookmark_name = name
            i = 0
            while bookmark_name in bookmarks:
                i += 1
                bookmark_name = "%s_%s" % (name, i)

            self.hg(["bookmark", bookmark_name])
            if not self.args.raw:
                logger.info("Bookmark set to %s", bookmark_name)

        if name and self.use_topic and config.create_topic and not self.args.no_topic:
            topics = self.hg_out(["topics", "-T", "{topic}\n"])
            topic_name = name
            i = 0
            while topic_name in topics:
                i += 1
                topic_name = "%s_%s" % (name, i)

            self.hg(["topic", topic_name])

    def apply_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ):
        # Import the diff to apply the changes then commit separately to
        # ensure correct parsing of the commit message.
        with temporary_file(diff) as diff_file, temporary_file(body) as msg_file:
            # Apply the patch with `--no-commit`, with file rename detection
            # (similarity). Using 95 as the similarity to match automv's default.
            self.hg(["import", "--no-commit", "-s", "95", diff_file, "--quiet"])

            commit_cmd = ["commit", "--logfile", msg_file]
            if author:
                commit_cmd.extend(["--user", author])
            if author_date:
                commit_cmd.extend(["--date", f"{author_date} 0"])

            self.hg(commit_cmd)

    def format_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ) -> str:
        changeset = ["# HG changeset patch"]
        if author:
            changeset.append("# User {}".format(author))
        if author_date:
            changeset.append("# Date {} 0".format(author_date))
        changeset.extend([body, diff])
        return "\n".join(changeset)

    def _amend_commit_body(self, node: str, body: str):
        with temporary_file(body) as body_file:
            self.checkout(node)
            self.hg(["commit", "--amend", "--logfile", body_file])

    def _get_parent(self, node: str) -> Optional[str]:
        return self.hg_out(
            ["log", "-T", "{node}", "-r", "parents(%s)" % node], split=False
        )

    def finalize(self, commits: List[Commit]):
        """Rebase stack children commits if needed."""
        # Currently we do all rebases in `amend_commit` if the evolve extension
        # is not installed.

        if not self.use_evolve:
            return

        parent = None
        for commit in commits:
            commit_parent = self._get_parent(commit.node)
            if parent and parent.node not in commit_parent:
                self.rebase_commit(commit, parent)
                (rev, node) = self._get_successor(commit.node)
                if rev and node:
                    self._refresh_commit(commit, node, rev)

            parent = commit

    def amend_commit(self, commit: Commit, commits: List[Commit]):
        updated_body = "%s\n%s" % (commit.title, commit.body)
        current_body = self.hg_out(
            ["log", "-T", "{desc}", "-r", commit.node], split=False
        )
        if current_body == updated_body:
            logger.debug("not amending commit %s, unchanged", commit.name)
            return

        # Find our position in the stack.
        parent_node = None
        first_child = None
        is_parent = True
        for stack_commit in commits:
            if stack_commit.node == commit.node:
                is_parent = False
            elif is_parent:
                parent_node = stack_commit.node
            elif not first_child:
                first_child = stack_commit
                break

        # Track children of this commit which aren't part of the stack.
        stack_nodes = [stack_commit.node for stack_commit in commits]
        non_stack_children = [
            n for n in self.hg_log(f"children({commit.node})") if n not in stack_nodes
        ]

        if self.use_evolve:
            # If evolve is installed this is trivial.
            self._amend_commit_body(commit.node, updated_body)

        elif not first_child and not non_stack_children:
            # Without evolve things are much more exciting.

            # If there's no children we can just amend.
            self._amend_commit_body(commit.node, updated_body)

            # This should always result in an amended node, but we need to be
            # extra careful not to strip the original node.
            amended_node = self.hg_log(".", split=False)
            if amended_node != commit.node:
                self.strip_nodes.append(commit.node)

        else:
            # Brace yourself.  We need to create a dummy commit with the same parent as
            # the commit, rebase a copy of the commit onto the dummy, amend, rebase the
            # amended commit back onto the original parent, rebase the children onto
            # that, then strip the original commit and dummy commits.

            # Find a parent for the first commit in the stack
            if not parent_node:
                parent_node = self.hg_log(f"parents({commit.node})")[0]

            # Create the dummy commit.
            self.checkout(parent_node)
            self.hg(
                ["commit"]
                + ["--message", "dummy"]
                + ["--config", "ui.allowemptycommit=true"]
            )
            dummy_node = self.hg_log(".", split=False)

            # Rebase a copy of this commit onto the dummy.
            self.hg(["rebase", "--keep", "--rev", commit.node, "--dest", dummy_node])
            rebased_node = self.hg_log("children(.)", split=False)

            # Amend.
            self._amend_commit_body(rebased_node, updated_body)
            amended_node = self.hg_log(".", split=False)

            # Rebase back onto parent
            self.hg(["rebase"] + ["--source", amended_node] + ["--dest", parent_node])
            rebased_amended_node = self.hg_log(".", split=False)

            # Update the commit object now.
            original_node = commit.node
            self._refresh_commit(commit, rebased_amended_node)

            # Note what nodes need to be stripped when we're all done.
            self.strip_nodes.extend([original_node, dummy_node])

            # And rebase children.
            if first_child:
                self.rebase_commit(first_child, commit)

        # Ensure our view of the stack is up to date.
        self.refresh_commit_stack(commits)

        # Commits that aren't part of the stack need to be re-parented.
        for node in non_stack_children:
            self.hg(["rebase", "--source", node, "--dest", commit.node])

    def is_descendant(self, node: str) -> bool:
        # Query the log for all commits that are both in the revset and descendants of
        # `node`. If we get any results, then our revset is already based off of `node`.
        out = self.hg_out(
            ["log", "-r", f"{node}:: and {self.revset}", "-T", "{node}\n"],
            split=False,
        )

        # If the command output is empty, our revset does not descend from `node`.
        if not out:
            return False

        return True

    def get_repo_head_branch(self) -> Optional[str]:
        shortname = self._phab_repo["fields"]["shortName"]
        if not self.is_node(shortname):
            return None

        return shortname

    def rebase_commit(self, source_commit: Commit, dest_commit: Commit):
        self.hg(
            ["rebase"] + ["--source", source_commit.node] + ["--dest", dest_commit.node]
        )

    def uplift_commits(self, dest: str, commits: List[Commit]) -> List[Commit]:
        try:
            out = self.hg_out(
                [
                    "rebase",
                    "-r",
                    self.revset,
                    "-d",
                    dest,
                    # Don't obsolete the original changesets.
                    "--keep",
                    "-T",
                    "json",
                    # Send messages to `stderr` so `stdout` is pure JSON.
                    "--config",
                    "ui.message-output=stderr",
                ],
                split=False,
            )
        except CommandError as exc:
            raise Error(
                f"Rebasing your uplift commits {self.revset} onto {dest} failed.\n\n"
                "This means your patch will fail to apply on landing due to conflicts "
                "with your desired uplift train.\n\n"
                f"Try rebasing with `hg rebase -r {self.revset} -d {dest} --keep`, "
                "resolving merge conflicts, and resubmitting."
            ) from exc

        # Capture the JSON output from rebase and use it to refresh our commit stack.
        # We need to do this since using `--keep` with rebase causes the newly produced
        # changesets to not be marked as successors of the original revisions, which
        # means `refresh_commit_stack` won't find any updates to make.
        nodechanges = json.loads(out)[0]["nodechanges"]

        if len(nodechanges) != len(commits):
            raise ValueError(
                f"Rebase created {len(nodechanges)} new changesets; "
                f"{len(commits)} expected"
            )

        for commit in commits:
            # We should have a single new commit for each commit we want to uplift.
            # Error out if we don't as progressing with an inconsistent stack might
            # cause behaviour we can't handle.
            new = nodechanges.get(commit.node)
            if len(new) != 1:
                raise ValueError(f"Should only have a single new item, got {new}")

            self._refresh_commit(commit, new[0])

        # Set revset to the new range of commits.
        self.revset = f"{commits[0].node}::{commits[-1].node}"

        return commits

    def check_commits_for_submit(self, commits: List[Commit]):
        # 'Greatest Common Ancestor'/'Merge Base' should be included in the revset.
        ancestor = self.hg_log("ancestor(%s)" % self.revset, split=False)
        if not any(commit.node == ancestor for commit in commits):
            raise Error(
                "Non-linear commit stack (common ancestor %s missing from stack)"
                % short_node(ancestor)
            )

        # Merge base needs to have a public parent.
        parent_phases = self.hg_out(
            ["log", "-T", "{phase} {node}\n", "-r", "parents(%s)" % ancestor]
        )
        for parent in parent_phases:
            (phase, node) = parent.split(" ", 1)
            if phase != "public":
                logger.warning(
                    "%s is based off non-public commit %s",
                    short_node(ancestor),
                    short_node(node),
                )

        # Can't submit merge requests.
        if self.hg_log("%s and merge()" % self.revset):
            raise Error("Commit stack contains a merge commit")

        # mq isn't currently supported.
        if self.has_mq and self.hg_out(["qapplied"]):
            raise Error("Found patches applied with `mq`, unable to continue")

        # Uncommitted changes can interact poorly when we update to a different commit.
        status = self._status()
        if status["T"]:
            err = [
                "%s uncommitted change%s present"
                % (len(status["T"]), " is" if len(status["T"]) == 1 else "s are")
            ]
            err.extend(
                [
                    "Commit changes, or use `hg shelve` to store uncommitted changes,",
                    "restoring with `hg unshelve` after submission",
                ]
            )
            if not self.has_shelve:
                err.append("You can enable the shelve extension via `hg config --edit`")
            raise Error("\n".join(err))

    def _get_file_modes(self, commit: Commit) -> dict:
        """Get modes of the modified files."""

        # build list of modified files
        # using a -T template here doesn't output all files (eg. source files from
        # a copy operation are skipped), so we have to parse the default output
        modified_files = [
            line[2:].replace("\\", "/")  # strip leading status char and space
            for line in self.hg_out(["status", "--change", commit.node, "--copies"])
        ]

        def _to_mode_dict(mode_list):
            mode_dict = {}
            for line in mode_list:
                flag, mode_path = line.split(":", 1)
                mode_dict[mode_path] = "100755" if flag == "x" else "100644"
            return mode_dict

        # get before/after file modes
        # .arcconfig is added to the file list to try to avoid the situation where
        # none of the files specified by -I are present in the commit.
        try:
            old_modes = _to_mode_dict(
                self.hg_out(
                    ["files"]
                    + ["--rev", commit.parent]
                    + ["-T", "{flags}:{path}\n"]
                    + ["-I%s" % f for f in modified_files]
                    + ["-I.arcconfig"]
                )
            )
        except CommandError:
            # Mercurial will return an error if none of the files on the -I list are
            # valid for the specified revision.  Treat any error here as an empty
            # result.
            old_modes = {}

        new_modes = _to_mode_dict(
            self.hg_out(
                ["files"]
                + ["--rev", commit.node]
                + ["-T", "{flags}:{path}\n"]
                + ["-I%s" % f for f in modified_files]
                + ["-I.arcconfig"]
            )
        )

        # build response
        file_modes = {}
        for path in modified_files:
            if path in old_modes:
                file_modes.setdefault(path, {})
                file_modes[path]["old_mode"] = old_modes[path]
            if path in new_modes:
                file_modes.setdefault(path, {})
                file_modes[path]["new_mode"] = new_modes[path]

        return file_modes

    def get_diff(self, commit: Commit) -> Diff:
        """Create a Diff object containing all changes for this commit."""
        self.checkout(commit.node)
        commit.parent = self._get_parent(commit.node)
        file_modes = self._get_file_modes(commit)

        # Get changed files.
        file_divider = "--%s--" % uuid.uuid4().hex
        type_divider = "--%s--" % uuid.uuid4().hex
        all_files = self.hg_out(
            ["log"]
            + ["-r", commit.node]
            + [
                "-T",
                "{{join(file_adds, '{file_divider}')}}{type_divider}"
                "{{join(file_dels, '{file_divider}')}}{type_divider}"
                "{{join(file_mods, '{file_divider}')}}{type_divider}"
                "{{join(file_copies, '{file_divider}')}}".format(
                    file_divider=file_divider, type_divider=type_divider
                ),
            ],
            split=False,
        )
        fn_adds, fn_dels, fn_mods, fn_renames_str = [
            fn.split(file_divider) for fn in all_files.split(type_divider)
        ]
        changes = []

        fn_renames = []
        fn_renamed = []  # collection of `old_fn`.
        # fn_renames_str is returning "new filename (old filename)"
        for fn in fn_renames_str:
            m = re.match(r"(?P<filename>[^(]+) \((?P<old_filename>[^)]+)\)", fn)
            if m:
                old_fn = m.group("old_filename")
                new_fn = m.group("filename")
                # A file can be mved only once.
                is_move = old_fn in fn_dels and old_fn not in fn_renamed
                changes.append(
                    {
                        "fn": new_fn,
                        "old_fn": old_fn,
                        "kind": "R" if is_move else "C",
                        "func": self._change_mod,
                    }
                )
                fn_renames.append((new_fn, old_fn))
                fn_renamed.append(old_fn)

        # remove renames from adds and dels
        fn_adds = [fn for fn in fn_adds if fn and fn not in [c[0] for c in fn_renames]]
        fn_dels = [fn for fn in fn_dels if fn and fn not in [c[1] for c in fn_renames]]

        # remove empty string from mods
        fn_mods = [fn for fn in fn_mods if fn]

        changes.extend(
            [{"fn": fn, "kind": "A", "func": self._change_add} for fn in fn_adds]
        )
        changes.extend(
            [{"fn": fn, "kind": "D", "func": self._change_del} for fn in fn_dels]
        )
        changes.extend(
            [{"fn": fn, "kind": "M", "func": self._change_mod} for fn in fn_mods]
        )

        # Create changes.
        diff = Diff()
        for c in changes:
            change = diff.change_for(c["fn"])
            old_fn = c["old_fn"] if "old_fn" in c else c["fn"]
            c["func"](change, c["fn"], old_fn, commit.parent, commit.node)
            a_mode = (
                file_modes[old_fn]["old_mode"]
                if old_fn in file_modes and "old_mode" in file_modes[old_fn]
                else "000000"
            )
            b_mode = (
                file_modes[c["fn"]]["new_mode"]
                if c["fn"] in file_modes and "new_mode" in file_modes[c["fn"]]
                else "000000"
            )
            diff.set_change_kind(change, c["kind"], a_mode, b_mode, old_fn, c["fn"])

        return diff

    @lru_cache(maxsize=None)  # noqa: B019
    def hg_cat(self, filename: str, node: str) -> Optional[bytes]:
        return self.hg_out(
            ["cat", "-r", node, filename], split=False, expect_binary=True
        )

    @lru_cache(maxsize=None)  # noqa: B019
    def _file_size(self, filename: str, rev: str) -> int:
        """Get the file size of the file."""
        return int(
            self.hg_out(
                [
                    "files",
                    "-v",
                    "-r",
                    rev,
                    os.path.join(self.path, filename),
                    "-T",
                    "{size}",
                ],
                split=False,
            )
        )

    @lru_cache(maxsize=128)  # noqa: B019
    def _get_file_meta(self, filename: str, rev: str) -> dict:
        """Collect information about the file."""
        binary = False
        body = self.hg_cat(filename, rev)
        meta = {"mime": "TEXT", "bin_body": body}

        meta["file_size"] = self._file_size(filename, rev)
        if meta["file_size"] > environment.MAX_TEXT_SIZE:
            binary = True

        if not binary:
            if b"\0" in body:
                binary = True

        if not binary:
            try:
                body = str(body, "utf-8")
            except UnicodeDecodeError:
                binary = True

        meta["body"] = body
        meta["binary"] = binary

        if binary:
            meta["mime"] = mimetypes.guess_type(filename)[0] or ""

        return meta

    def _change_add(
        self,
        change: Diff.Change,
        filename: str,
        _old_filename: str,
        _parent: str,
        node: str,
    ):
        """Create a change about adding a file to the commit."""
        meta = self._get_file_meta(filename, node)
        telemetry().submission.files_size.accumulate(meta["file_size"])

        if meta["binary"]:
            change.set_as_binary(
                a_body="",
                a_mime="",
                b_body=meta["bin_body"],
                b_mime=meta["mime"],
            )
            return

        # Empty files don't have any hunks.
        if meta["file_size"] == 0:
            return

        lines, eof_missing_newline = create_hunk_lines(meta["body"], "+")
        new_len = len(lines)
        if eof_missing_newline:
            new_len -= 1

        change.hunks.append(
            Diff.Hunk(
                old_off=0,
                new_off=1,
                old_len=0,
                new_len=new_len,
                lines=lines,
            )
        )

    def _change_del(
        self,
        change: Diff.Change,
        filename: str,
        _old_filename: str,
        parent: str,
        _node: str,
    ):
        """Create a change about deleting a file from the commit."""
        meta = self._get_file_meta(filename, parent)
        telemetry().submission.files_size.accumulate(meta["file_size"])

        if meta["binary"]:
            change.set_as_binary(
                a_body=meta["bin_body"],
                a_mime=meta["mime"],
                b_body="",
                b_mime="",
            )
            return

        # Empty files don't have any hunks.
        if meta["file_size"] == 0:
            return

        lines = create_hunk_lines(meta["body"], "-")[0]
        old_len = len(lines)

        change.hunks.append(
            Diff.Hunk(
                old_off=1,
                new_off=0,
                old_len=old_len,
                new_len=0,
                lines=lines,
            )
        )

    def _change_mod(
        self,
        change: Diff.Change,
        filename: str,
        old_filename: str,
        parent: str,
        node: str,
    ):
        """Create a change about modified file in the commit."""
        a_meta = self._get_file_meta(old_filename, parent)
        b_meta = self._get_file_meta(filename, node)
        file_size = max(a_meta["file_size"], b_meta["file_size"])
        telemetry().submission.files_size.accumulate(file_size)

        if a_meta["binary"] or b_meta["binary"]:
            # Binary file.
            change.set_as_binary(
                a_body=a_meta["bin_body"],
                a_mime=a_meta["mime"],
                b_body=b_meta["bin_body"],
                b_mime=b_meta["mime"],
            )
            return

        # Empty files don't have any hunks.
        if a_meta["file_size"] == 0 and b_meta["file_size"] == 0:
            return

        if a_meta["body"] == b_meta["body"]:
            # File contents unchanged.
            lines = create_hunk_lines(a_meta["body"], " ", check_eof=False)[0]
            change.hunks.append(
                Diff.Hunk(
                    old_off=1,
                    new_off=1,
                    old_len=len(lines),
                    new_len=len(lines),
                    lines=lines,
                )
            )
            return

        if self.args.lesscontext or file_size > environment.MAX_CONTEXT_SIZE:
            context_size = 100
        else:
            context_size = environment.MAX_CONTEXT_SIZE

        start = time.process_time()

        # Using binary here to ensure we avoid converting newlines
        git_diff = self.hg_out(
            [
                "diff",
                "--git",
                "--unified",
                str(context_size),
                "--rev",
                parent,
                filename,
            ],
            expect_binary=True,
        ).decode("utf-8")
        change.from_git_diff(git_diff)

        if file_size > environment.MAX_CONTEXT_SIZE / 2:
            logger.debug(
                "Splitting the diff (size: %s) took %ss",
                file_size,
                time.process_time() - start,
            )

    def validate_email(self):
        email = self.extract_email_from_username()
        if not is_valid_email(email):
            raise Error(
                f"Your username configured with Mercurial ({email}) "
                f"must contain a valid email.\n"
                f"Please see https://www.mercurial-scm.org/doc/hgrc.5.html "
                f"for more information on editing your Mercurial configuration."
                f"\n\nYou can also amend a commit via "
                f'`hg commit --amend --user "username <someone@example.com>"'
            )

    def extract_email_from_username(self) -> str:
        """Extracts an email from a Mercurial username, if it exists.
        Not guaranteed to return a valid email, make sure to validate."""
        return self.username.split("<").pop().replace(">", "")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import platform
from pathlib import Path
from typing import (
    Union,
)

import distro

from mozphab import environment

from .bmo import BMOAPIError
from .config import config
from .environment import MOZPHAB_VERSION
from .helpers import prompt
from .logger import logger
from .user import user_data


class Telemetry:
    def __init__(self):
        """Initiate Glean, load pings and metrics."""
        import glean

        logging.getLogger("glean").setLevel(logging.DEBUG)
        logger.debug("Initializing Glean...")

        glean.Glean.initialize(
            application_id="MozPhab",
            application_version=MOZPHAB_VERSION,
            upload_enabled=True,
            configuration=glean.Configuration(),
            data_dir=Path(environment.MOZBUILD_PATH) / "telemetry-data",
        )

        self._pings = glean.load_pings(environment.MOZPHAB_MAIN_DIR / "pings.yaml")
        self._metrics = glean.load_metrics(
            environment.MOZPHAB_MAIN_DIR / "metrics.yaml"
        )

    @property
    def environment(self):
        return self._metrics.mozphab.environment

    @property
    def usage(self):
        return self._metrics.mozphab.usage

    @property
    def user(self):
        return self._metrics.mozphab.user

    @property
    def submission(self):
        return self._metrics.mozphab.submission

    def _set_os(self):
        """Collect human readable information about the OS version.

        For Linux it is setting a distribution name and version.
        """
        system, node, release, version, machine, processor = platform.uname()
        if system == "Linux":
            distribution_name, distribution_number, _ = distro.linux_distribution(
                full_distribution_name=False
            )
            distribution_version = " ".join([distribution_name, distribution_number])
        elif system == "Windows":
            _release, distribution_version, _csd, _ptype = platform.win32_ver()
        elif system == "Darwin":
            distribution_version, _versioninfo, _machine = platform.mac_ver()
        else:
            distribution_version = release

        self.environment.distribution_version.set(distribution_version)

    def _set_python(self):
        self.environment.python_version.set(platform.python_version())

    def set_vcs(self, repo):
        self.environment.vcs.name.set(repo.vcs)
        self.environment.vcs.version.set(repo.vcs_version)

    def submit(self):
        self._pings.usage.submit()
        logger.debug("Telemetry submit called.")

    def set_metrics(self, args):
        """Sets metrics common to all commands."""
        self.usage.command.set(args.command)
        self._set_os()
        self._set_python()
        self.usage.override_switch.set(
            getattr(args, "force_vcs", False) or getattr(args, "force", False)
        )
        self.usage.command_time.start()
        self.user.installation.set(user_data.installation_id)
        self.user.id.set(user_data.user_code)


class TelemetryDisabled:
    """Dummy class that does nothing."""

    def __init__(*args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, *args, **kwargs):
        return self


def update_user_data():
    """Update user_data to enable or disable Telemetry.

    If employment data has been changed Telemetry might be switched on
    automatically. The opt-in decision is taken for the new employee. Non employees
    will have an option to enable data collection.
    """
    is_employee_changed = user_data.set_user_data()
    if not is_employee_changed:
        return

    # Switch on Telemetry for employee or ask to opt-in for non-employee
    if user_data.is_employee:
        logger.warning(
            "Enabled collecting MozPhab usage data.\n"
            "See https://moz-conduit.readthedocs.io/en/latest"
            "/mozphab-data-collection.html"
        )
        config.telemetry_enabled = True
    else:
        # user is new or no longer employee
        opt_in = (
            prompt(
                "Would you like to allow MozPhab to collect usage data?",
                ["Yes", "No"],
            )
            == "Yes"
        )
        if opt_in:
            config.telemetry_enabled = True
        else:
            logger.info(
                "MozPhab usage data collection disabled.\n"
                "See https://moz-conduit.readthedocs.io/en/latest"
                "/mozphab-data-collection.html"
            )
            config.telemetry_enabled = False
    config.write()


def configure_telemetry(args):
    if args.command == "install-certificate":
        # Collecting data without a certificate is not possible.
        _Globals.telemetry = TelemetryDisabled()
        return

    if args.command == "self-update":
        # Avoid locking issues on Windows by not loading Glean when we're updating
        _Globals.telemetry = TelemetryDisabled()
        return

    # `user_data` file will remain empty until user calls MozPhab with a command
    # requiring existence of the Repository.
    if args.needs_repo:
        try:
            update_user_data()
        except BMOAPIError as err:
            # Error in retrieving user status.
            # We quietly allow to work without enabling Telemetry.
            logger.debug("BMOAPIError: %s", err)
            _Globals.telemetry = TelemetryDisabled()
            return

    # We can't call telemetry if user data was never collected.
    if not config.telemetry_enabled or not user_data.is_data_collected:
        _Globals.telemetry = TelemetryDisabled()
        return

    # Enable telemetry by swapping the telemetry global with a Glean backed object.
    _Globals.telemetry = Telemetry()
    telemetry().set_metrics(args)


def telemetry():
    return _Globals.telemetry


class _Globals:
    """Container for holding globals in a way that can be easily replaced."""

    telemetry: Union[Telemetry, TelemetryDisabled] = TelemetryDisabled()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import io
import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import uuid
from pathlib import Path
from unittest import mock

import pytest
from glean import testing

# NOTE: Set the environment variables before importing any of mozphab modules.
os.environ["MOZPHAB_NO_USER_CONFIG"] = "1"

from mozphab import (  # noqa: E402
    conduit,
    environment,
    mozphab,
    repository,
    simplecache,
    user,
)
from mozphab import (  # noqa: E402
    config as config_module,
)
from mozphab.commands import submit  # noqa: E402
from mozphab.git import Git  # noqa: E402
from mozphab.mercurial import Mercurial  # noqa: E402

environment.SHOW_SPINNER = False
environment.HTTP_ALLOWED = True


def create_temp_fn(*filenames):
    m_temp_fn = mock.Mock()
    if len(filenames) > 1:
        m_temp_fn.__enter__ = mock.Mock(side_effect=filenames)
    else:
        m_temp_fn.__enter__ = mock.Mock(return_value=filenames[0])
    m_temp_fn.__exit__ = mock.Mock(return_value=None)
    return m_temp_fn


def search_diff(diff=1, phid="PHID-DIFF-1", node="aaa000aaa000"):
    return {
        "id": diff,
        "phid": phid,
        "attachments": {"commits": {"commits": [{"identifier": node}]}},
    }


def search_rev(
    rev=1,
    phid="PHID-DREV-1",
    bug="1",
    status="needs-review",
    closed=False,
    author="PHID-USER-1",
    diff="PHID-DIFF-1",
    repo="PHID-REPO-1",
    reviewers=None,
    is_draft=False,
):
    reviewers = reviewers or []
    return {
        "id": rev,
        "phid": phid,
        "fields": {
            "bugzilla.bug-id": bug,
            "status": {"value": status, "closed": closed},
            "authorPHID": author,
            "diffPHID": diff,
            "repositoryPHID": repo,
            "isDraft": is_draft,
        },
        "attachments": {
            "reviewers": {"reviewers": [{"reviewerPHID": r} for r in reviewers]}
        },
    }


def assert_attributes(hunk, expected):
    for name, value in expected.items():
        if hunk.__getattribute__(name) != value:
            raise AssertionError(
                "%s: expected '%s', got '%s'"
                % (name, value, hunk.__getattribute__(name))
            )


@pytest.fixture
def posix_linesep():
    """POSIX line separator (i.e. newline character)."""
    return "\n"


@pytest.fixture
def hg_sha():
    def ret():
        return hg_out("id", "-i").rstrip("\n")

    return ret


@pytest.fixture
def git_sha():
    def ret():
        return git_out("log", "--format=%H", "-1").rstrip("\n")

    return ret


@pytest.fixture
def init_sha(in_process, git_repo_path, git_sha):
    return git_sha()


@pytest.fixture(autouse=True)
def reset_cache():
    simplecache.cache.reset()


@pytest.fixture()
def repo_phab_url():
    with mock.patch("mozphab.repository.Repository._phab_url") as xmock:
        xmock.return_value = "http://phab.test"
        yield xmock


@pytest.fixture
def data_file():
    return Path(__file__).parent / "data" / "img.png"


@pytest.fixture
def git_command():
    mozphab.config.git_command = ["git"]
    return mozphab.config


@pytest.fixture
@mock.patch("mozphab.gitcommand.GitCommand.output")
@mock.patch("mozphab.git.Git._get_current_head")
@mock.patch("mozphab.repository.os.path")
@mock.patch("mozphab.helpers.which")
@mock.patch("mozphab.repository.read_json_field")
def git(
    m_read_json_field,
    m_which,
    m_os_path,
    m_git_get_current_head,
    m_git_git_out,
    repo_phab_url,
    git_command,
):
    m_read_json_field.return_value = "TEST"
    m_os_path.join = os.path.join
    m_os_path.exists.return_value = True
    m_which.return_value = True
    m_os_path.isfile.return_value = False
    m_git_git_out.side_effect = ("git version 2.25.0",)
    m_git_get_current_head.return_value = "branch"
    git = Git("x")
    git._phab_vcs = "git"
    git._phab_repo = {
        "fields": {
            "defaultBranch": "beta",
            "shortName": "firefox-beta",
            "vcs": "git",
        }
    }
    git.args = argparse.Namespace(
        upstream=[],
        safe_mode=False,
        no_branch=False,
        yes=False,
        lesscontext=False,
        force_vcs=False,
        command="submit",
    )
    return git


@pytest.fixture
@mock.patch("mozphab.repository.os.chdir")
@mock.patch("mozphab.mercurial.Mercurial.repository")
@mock.patch("mozphab.repository.os.path")
@mock.patch("mozphab.helpers.which")
@mock.patch("mozphab.repository.read_json_field")
@mock.patch("mozphab.mercurial.hglib.open")
def hg(
    m_hglib_open,
    m_read_json_field,
    m_which,
    m_os_path,
    m_repository,
    m_os_chdir,
    safe_environ,
    repo_phab_url,
):
    m_read_json_field.return_value = "TEST"
    m_os_path.join = os.path.join
    m_os_path.exists.return_value = True
    m_which.return_value = True
    m_os_path.isfile.return_value = False
    m_repository.version = 4, 7, 1, "bleh"
    m_os_chdir.return_value = True
    mozphab.config.hg_command = ["hg"]
    hg = Mercurial("x")
    hg.use_evolve = True
    hg.has_mq = False
    hg._phab_vcs = "hg"
    hg._phab_repo = {
        "fields": {
            "defaultBranch": "default",
            "shortName": "beta",
            "vcs": "hg",
        }
    }
    return hg


def hg_out(*args):
    args = ["hg"] + list(args)
    return subprocess.check_output(args, universal_newlines=True, encoding="utf-8")


def write_text(filename, content, **kwargs):
    # Always write LF delimited text
    with io.open(filename, "w", newline="\n", **kwargs) as f:
        f.write(content)


def find_script_path(name):
    """Return the fully qualified path to an executable, preferring those installed
    into the current virtualenv."""
    bin_path = Path(sys.executable).parent
    if sys.platform == "win32":
        name = "%s.exe" % name
    filename = bin_path / name
    if filename.exists():
        return str(filename)
    # Check system path
    if shutil.which(name):
        return name
    raise Exception("Failed to find script: %s" % filename)


@pytest.fixture
def hg_repo_path(monkeypatch, tmp_path):
    """Build a usable HG repository. Return the pathlib.Path to the repo."""
    phabricator_uri = "http://example.test"
    repo_path = tmp_path / "hg-repo"
    repo_path.mkdir()
    monkeypatch.chdir(str(repo_path))
    arcconfig = repo_path / ".arcconfig"
    arcconfig.write_text(json.dumps({"phabricator.uri": phabricator_uri}))
    hg_out("init")
    hg_out("add")
    hg_out("commit", "-m", "init")
    # graphshorten changes `log --graph` output, force to false
    with open(".hg/hgrc", "a") as f:
        f.write("\n[experimental]\ngraphshorten = false\n")
    return repo_path


@pytest.fixture
def fresh_global_config(tmp_path):
    """Overrides global ~/.gitconfig.

    Creates a tiny gitconfig file in the temp repo and sets it as the $HOME
    """
    original_env = os.environ.copy()
    env = os.environ
    env["HOME"] = str(tmp_path)
    with open("{}/.gitconfig".format(tmp_path), "w") as f:
        f.write("[user]\n\tname = Developer\n\temail = developer@mozilla.com\n")
    with open("{}/.hgrc".format(tmp_path), "w") as f:
        f.write(
            "[ui]\nusername = Developer <developer@mozilla.com>\n"
            "[extensions]\nevolve =\n"
        )
    yield
    os.environ.clear()
    os.environ.update(original_env)


def git_out(*args):
    env = os.environ.copy()
    args = [
        "git",
        "-c",
        "i18n.logOutputEncoding=UTF-8",
        "-c",
        "i18n.commitEncoding=UTF-8",
    ] + list(args)
    env["DEBUG"] = "1"
    return subprocess.check_output(
        args, env=env, universal_newlines=True, encoding="utf-8"
    )


def jj_out(*args):
    env = os.environ.copy()
    args = ["jj", "--debug"] + list(args)
    return subprocess.check_output(
        args, env=env, universal_newlines=True, encoding="utf-8"
    )


@pytest.fixture
def git_repo_path(monkeypatch, tmp_path):
    """Build a usable Git repository. Return the pathlib.Path to the repo."""
    phabricator_uri = "http://example.test"
    repo_path = tmp_path / "git-repo"
    repo_path.mkdir()
    monkeypatch.chdir(str(repo_path))
    arcconfig = repo_path / ".arcconfig"
    arcconfig.write_text(json.dumps({"phabricator.uri": phabricator_uri}))

    # Use --initial-branch where available to avoid an unnecessary warning
    m = re.search(r"(\d+\.\d+)\.\d+", git_out("version"))
    if m and float(m[1]) >= 2.28:
        git_out("init", "--initial-branch", "main")
    else:
        git_out("init")

    git_out("add", ".")
    git_out("commit", "--message", "initial commit")
    return repo_path


@pytest.fixture
def jj_colocated_repo_path(monkeypatch, tmp_path):
    """Build a usable Jujutsu repository. Return the pathlib.Path to the repo."""
    phabricator_uri = "http://example.test"
    repo_path = tmp_path / "jj-repo"
    repo_path.mkdir()
    monkeypatch.chdir(str(repo_path))
    arcconfig = repo_path / ".arcconfig"
    arcconfig.write_text(json.dumps({"phabricator.uri": phabricator_uri}))

    jj_out("git", "init", "--colocate")
    jj_out("commit", "--message", "initial commit")
    return repo_path


@pytest.fixture
def safe_environ(monkeypatch):
    # Make sure we preserve the system defaults.
    monkeypatch.setattr(os, "environ", os.environ.copy())
    # Disable logging to keep the testrunner output clean
    monkeypatch.setattr(mozphab, "init_logging", mock.MagicMock())


@pytest.fixture
def config():
    """Replace the original config with the fake one."""
    with tempfile.NamedTemporaryFile() as temp:
        return config_module.Config(filename=Path(temp.name))


@pytest.fixture
def in_process(monkeypatch, safe_environ, request, config):
    """Set up an environment to run moz-phab within the current process."""
    monkeypatch.setattr("mozphab.config.config", config)
    monkeypatch.setattr("mozphab.git.config", config)
    monkeypatch.setattr("mozphab.mercurial.config", config)
    # Make sure other tests didn't leak and mess up the module-level
    # global variables :/
    monkeypatch.setattr(environment, "DEBUG", True)
    monkeypatch.setattr(environment, "HAS_ANSI", False)
    # Constructing the Mercurial() object modifies os.environ for all tests.
    # Disable update checking.  It modifies the program on disk which we do /not/ want
    # to do during a test run.
    monkeypatch.setattr(mozphab, "check_for_updates", mock.Mock(return_value=None))

    # Disable calls to sys.exit() at the end of the script.  Re-raise errors instead
    # to make test debugging easier.
    def reraise(*args, **kwargs):
        raise

    monkeypatch.setattr(sys, "exit", reraise)

    # Disable uploading a new commit title and summary to Phabricator.  This operation
    # is safe to skip and doing so makes it easier to test other conduit call sites.
    monkeypatch.setattr(submit, "update_revision_description", mock.MagicMock())

    # Modify user_data object to not touch the file
    user.USER_INFO_FILE = mock.Mock()
    user.USER_INFO_FILE.exists.return_value = False
    user.user_data = user.UserData()
    user.user_data.update_from_dict(
        {
            "user_code": str(uuid.uuid4()),
            "is_employee": True,
            "installation_id": str(uuid.uuid4()),
            "last_check": time.time(),
        }
    )

    # Allow to define the check_call_by_line function in the testing module
    def check_call_by_line_static(*args, **kwargs):
        return ["Revision URI: http://example.test/D123"]

    def call_conduit_static(self, *args):
        # Return alice as the only valid reviewer name from Phabricator.
        # See https://phabricator.services.mozilla.com/api/user.search
        return [{"userName": "alice"}]

    call_conduit = getattr(request.module, "call_conduit", call_conduit_static)
    monkeypatch.setattr(conduit.ConduitAPI, "call", call_conduit)

    def read_json_field_local(self, *args):
        if args[0][0] == "phabricator.uri":
            return "http://example.test"
        elif args[0][0] == "repository.callsign":
            return "TEST"

    read_json_field = getattr(request.module, "read_json_field", read_json_field_local)
    monkeypatch.setattr(repository, "read_json_field", read_json_field)


@pytest.fixture
@mock.patch("mozphab.user.USER_INFO_FILE")
def user_data(m_file):
    m_file.exists.return_value = False
    return user.UserData()


@pytest.fixture(name="reset_glean", scope="function", autouse=True)
def fixture_reset_glean():
    testing.reset_glean(application_id="mozphab", application_version="0.1.86")


@pytest.fixture(autouse=True)
def mock_load_api_token(monkeypatch, request):
    if "no_mock_token" not in request.keywords:
        monkeypatch.setattr("mozphab.conduit.ConduitAPI.load_api_token", mock.MagicMock)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import logging
import textwrap
from typing import List

from mozphab import environment
from mozphab.commits import Commit
from mozphab.conduit import conduit
from mozphab.config import config
from mozphab.exceptions import Error
from mozphab.helpers import (
    BLOCKING_REVIEWERS_RE,
    augment_commits_from_body,
    has_arc_rejections,
    move_drev_to_original,
    prompt,
    strip_differential_revision,
    strip_dontbuild,
    update_commit_title_previews,
)
from mozphab.logger import logger
from mozphab.repository import Repository
from mozphab.spinner import wait_message
from mozphab.telemetry import telemetry


def morph_blocking_reviewers(commits: List[Commit]):
    """Automatically fix common typo by replacing r!user with r=user!"""

    def morph_reviewer(matchobj):
        if matchobj.group(1) == "r!":
            nick = matchobj.group(2)

            # strip trailing , or . so we can put it back later
            if nick[-1] in (",", "."):
                suffix = nick[-1]
                nick = nick[:-1]
            else:
                suffix = ""

            # split on comma to support r!a,b -> r=a!,b!
            return "r=%s%s" % (
                ",".join(["%s!" % n.rstrip("!") for n in nick.split(",")]),
                suffix,
            )
        else:
            return matchobj.group(0)

    for commit in commits:
        commit.title = BLOCKING_REVIEWERS_RE.sub(morph_reviewer, commit.title)


def amend_revision_url(body: str, new_url: str) -> str:
    """Append or replace the Differential Revision URL in a commit body."""
    body = strip_differential_revision(body)
    if body:
        body += "\n"
    body += "\nDifferential Revision: %s" % new_url
    return body


def log_commit_stack_with_messages(
    commits: List[Commit],
    messages=None,
    initial_indent="",
    subsequent_indent="",
    level=logging.WARNING,
):
    """Log the commit stack in a human readable form."""
    messages = messages or {}

    # Keep output in columns by sizing the action column to the longest rev ID.
    max_len = max(
        (len(f"(D{commit.rev_id})") for commit in commits if commit.rev_id), default=""
    )
    action_template = f"%{max_len}s %s %s"

    for commit in reversed(commits):
        action = f"(D{commit.rev_id})" if commit.rev_id else "(New)"
        logger.info(action_template, action, commit.name, commit.revision_title())
        for message in messages.get(commit.name, []):
            logger.log(
                level,
                textwrap.fill(
                    message,
                    initial_indent=initial_indent,
                    subsequent_indent=subsequent_indent,
                ),
            )


def show_commit_stack(commits: List[Commit]):
    """Show the commits that were submitted and their URLs."""
    submitted_commits = []
    urls = {}

    for commit in commits:
        if not commit.rev_id or not commit.submit:
            continue
        submitted_commits.append(commit)
        urls[commit.name] = [f"{conduit.repo.phab_url}/D{commit.rev_id}"]

    log_commit_stack_with_messages(submitted_commits, urls, "-> ")


def validate_commit_stack(
    commits: List[Commit], args: argparse.Namespace
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    """Validate commit stack is suitable for review.

    Return collections of warnings and errors.
    """

    # Preload all revisions.
    ids = [commit.rev_id for commit in commits if commit.rev_id]
    with wait_message("Loading existing revisions..."):
        revisions = conduit.get_revisions(ids=ids) if ids else []
        revisions = {revision["id"]: revision for revision in revisions}

    # Preload diffs.
    with wait_message("Loading diffs..."):
        diff_phids = [revision["fields"]["diffPHID"] for revision in revisions.values()]
        diffs = conduit.get_diffs(phids=diff_phids) if diff_phids else {}

    warnings = {}
    errors = {}
    nodes = {}

    for commit in commits:
        if revision := revisions.get(commit.rev_id):
            if commit.name != (dupe := nodes.setdefault(commit.rev_id, commit.name)):
                errors.setdefault(commit.name, []).append(
                    f"Phabricator revisions should be unique, but commit {dupe} refers "
                    f"to the same one D{commit.rev_id}."
                )
                # Don't show more for this commit because it's probably wrong.
                continue

            commit.rev_phid = revision["phid"]
            fields = revision["fields"]

            # WIP if either in changes-planned state, or if the revision has the
            # `draft` flag set.
            revision_is_wip = bool(
                fields["status"]["value"] == "changes-planned" or fields["isDraft"]
            )

            # Check if target bug ID is the same as in the Phabricator revision.
            bug_id_changed = bool(
                fields.get("bugzilla.bug-id")
                and commit.bug_id != fields["bugzilla.bug-id"]
            )

            # Check if revision is closed.
            revision_is_closed = bool(fields["status"]["closed"])

            # Check if comandeering is required.
            whoami = conduit.whoami()
            different_author = fields["authorPHID"] != whoami["phid"]

            # Any reviewers added to a revision without them?
            reviewers_added = bool(
                commit.reviewers["granted"]
                and not revision["attachments"]["reviewers"]["reviewers"]
            )

            # If SHA1 hasn't changed
            # and we're not changing the WIP or draft status
            # and we're not adding reviewers to a revision without reviewers
            # and we're not changing the bug ID
            # then don't submit.
            diff_phid = fields["diffPHID"]
            diff_commits = diffs[diff_phid]["attachments"]["commits"]["commits"]
            sha1_changed = bool(
                not diff_commits or commit.node != diff_commits[0]["identifier"]
            )
            if (
                not sha1_changed
                and commit.wip == revision_is_wip
                and not reviewers_added
                and not bug_id_changed
                and not revision_is_closed
                and not args.force
            ):
                commit.submit = False
                warnings.setdefault(commit.name, []).append(
                    "This revision has not changed and will not be submitted."
                )
                continue

            if not commit.wip and revision_is_wip:
                warnings.setdefault(commit.name, []).append(
                    '"Changes Planned" status will change to "Request Review"'
                )
            elif commit.wip and not revision_is_wip:
                warnings.setdefault(commit.name, []).append(
                    '"Request Review" status will change to "Changes Planned"'
                )
            if bug_id_changed:
                warnings.setdefault(commit.name, []).append(
                    "Bug ID in Phabricator revision will change from "
                    f"{fields['bugzilla.bug-id']} to {commit.bug_id}"
                )
            if different_author:
                warnings.setdefault(commit.name, []).append(
                    "You don't own this revision. Normally, you should only update "
                    'revisions you own. You can "Commandeer" this revision from the '
                    "web interface if you want to become the owner."
                )
            if revision_is_closed:
                warnings.setdefault(commit.name, []).append(
                    "This revision is closed! It will be reopened if submission "
                    "proceeds. You can stop now and refine the stack range."
                )
        elif commit.rev_id:
            errors.setdefault(commit.name, []).append(
                "Phabricator didn't return a query result for revision "
                f"D{commit.rev_id}. It might be inaccessible or not exist at all."
            )
            # Unrecoverable error for this commit.
            continue

        overridable_errors = []

        if not commit.bug_id and not args.no_bug:
            overridable_errors.append("Missing bug ID")

        if has_arc_rejections(commit.body):
            overridable_errors.append("Contains arc fields")

        if (
            commit.has_reviewers
            and not commit.wip
            and not conduit.has_revision_reviewers(commit)
            and (invalid := conduit.check_for_invalid_reviewers(commit.reviewers))
        ):
            for reviewer in invalid:
                if "disabled" in reviewer:
                    message = f"User {reviewer['name']} is disabled"
                elif "until" in reviewer:
                    message = (
                        f"{reviewer['name']} isn't available until {reviewer['until']} "
                        "(override with `-f`)"
                    )
                else:
                    message = f"{reviewer['name']} isn't a valid reviewer name"
                overridable_errors.append(message)

        if overridable_errors:
            if args.force:
                # Downgrade to warnings if `-f` was passed.
                warnings.setdefault(commit.name, []).extend(overridable_errors)
            else:
                errors.setdefault(commit.name, []).extend(overridable_errors)

        if commit.bug_id_orig and commit.bug_id != commit.bug_id_orig:
            warnings.setdefault(commit.name, []).append(
                f"Bug ID changed from {commit.bug_id_orig} to {commit.bug_id}"
            )

        if (
            not commit.has_reviewers
            and not conduit.has_revision_reviewers(commit)
            # Submitting a new uplift clears reviewers and shouldn't warn.
            and args.command != "uplift"
        ):
            warnings.setdefault(commit.name, []).append("Missing reviewers")

        if commit.wip and not args.wip:
            warnings.setdefault(commit.name, []).append(
                'It will be submitted as "Changes Planned". Run submit again with '
                "`--no-wip` to prevent this."
            )

    return warnings, errors


def make_blocking(reviewers: List[str]) -> List[str]:
    """Convert a list of reviewer strings into a list of blocking reviewer strings."""
    return ["%s!" % r.rstrip("!") for r in reviewers]


def remove_duplicates(reviewers: List[str]) -> List[str]:
    """Remove all duplicate items from the list.

    Args:
        reviewers: list of strings representing IRC nicks of reviewers.

    Returns: list of unique reviewers.

    Duplicates with excalamation mark are preferred.
    """
    unique = []
    nicks = []
    for reviewer in reviewers:
        nick = reviewer.lower().strip("!")
        if reviewer.endswith("!") and nick in nicks:
            nicks.remove(nick)
            unique = [r for r in unique if r.lower().strip("!") != nick]

        if nick not in nicks:
            nicks.append(nick)
            unique.append(reviewer)

    return unique


def update_commits_from_args(commits: List[Commit], args: argparse.Namespace):
    """Modify commit description based on args and configuration.

    Args:
        commits: list of Commits representing the commits in commit stack
        args: argparse.Namespace object. In this function we're using
            following attributes:
            * reviewer - list of strings representing reviewers
            * blocker - list of string representing blocking reviewers
            * bug - an integer representing bug id in Bugzilla

    Command args always overwrite commit desc.
    Overwriting reviewers rules
    (The "-r" are loaded into args.reviewer and "-R" into args.blocker):
        a) Recognize `r=` and `r?` from commit message title.
        b) Modify reviewers only if `-r` or `-R` is used in call arguments.
        c) Add new reviewers using `r=`.
        d) Do not modify reviewers already provided in the title (`r?` or `r=`).
        e) Delete reviewer if `-r` or `-R` is used and the reviewer is not mentioned.

    If no reviewers are provided by args and an `always_blocking` flag is set
    change all reviewers in title to blocking.
    """

    # Build list of reviewers from args.
    reviewers = list(set(args.reviewer)) if args.reviewer else []

    # Order might be changed here `list(set(["one", "two"])) == ['two', 'one']`
    reviewers.sort()
    blockers = [r.rstrip("!") for r in args.blocker] if args.blocker else []

    # User might use "-r <nick>!" to provide a blocking reviewer.
    # Add all such blocking reviewers to blockers list.
    blockers += [r.rstrip("!") for r in reviewers if r.endswith("!")]
    blockers = list(set(blockers))
    blockers.sort()

    # Remove blocking reviewers from reviewers list
    reviewers_no_flag = [r.strip("!") for r in reviewers]
    reviewers = [r for r in reviewers_no_flag if r.lower() not in blockers]

    # Add all blockers to reviewers list. Reviewers list will contain a list
    # of all reviewers where blocking ones are marked with the "!" suffix.
    reviewers += make_blocking(blockers)
    reviewers = remove_duplicates(reviewers)
    lowercase_reviewers = [r.lower() for r in reviewers]
    lowercase_blockers = [r.lower() for r in blockers]

    for commit in commits:
        if reviewers:
            # Only the reviewers mentioned in command line args will remain.
            # New reviewers will be marked as "granted" (r=).
            granted = reviewers.copy()
            requested = []

            # commit["reviewers"]["request|"] is a list containing reviewers
            # provided in commit's title with an r? mark.
            for reviewer in commit.reviewers["request"]:
                r = reviewer.strip("!")
                # check which request reviewers are provided also via -r
                if r.lower() in lowercase_reviewers:
                    requested.append(r)
                    granted.remove(r)
                # check which request reviewers are provided also via -R
                elif r.lower() in lowercase_blockers:
                    requested.append("%s!" % r)
                    granted.remove("%s!" % r)
        else:
            granted = commit.reviewers.get("granted", [])
            requested = commit.reviewers.get("request", [])

        commit.reviewers = {"granted": granted, "request": requested}

        if args.bug:
            # Bug ID command arg used.
            commit.bug_id = args.bug

        # Mark a commit as WIP if --wip is provided or if the revision has no reviewers.
        # This is after checking for WIP: prefix in helpers.augment_commits_from_body().
        if args.no_wip:
            commit.wip = False
        elif args.wip or (
            # `moz-phab uplift` assumes `#release-managers` review will be set
            # by automation once the assessment form is submitted.
            args.command != "uplift"
            and not commit.has_reviewers
            and not conduit.has_revision_reviewers(commit)
        ):
            commit.wip = True
        elif commit.wip is None:
            commit.wip = False

    # Honour config setting to always use blockers
    if not reviewers and config.always_blocking:
        for commit in commits:
            commit.reviewers = {
                "request": make_blocking(commit.reviewers["request"]),
                "granted": make_blocking(commit.reviewers["granted"]),
            }


def update_commits_for_uplift(commits: List[Commit], repo: Repository):
    """Prepares a set of commits for uplifting."""
    with wait_message("Loading revision data..."):
        rev_ids = [commit.rev_id for commit in commits if commit.rev_id]
        revisions = conduit.get_revisions(ids=rev_ids) if rev_ids else []
        revisions = {revision["id"]: revision for revision in revisions}

    for commit in commits:
        # Clear all reviewers from the revision.
        commit.reviewers = {
            "granted": [],
            "request": [],
        }

        # Remove DONTBUILD flag from commit title on uplift.
        commit.title = strip_dontbuild(commit.title)

        # Skip when updating an existing revision on the uplift repo.
        revision = revisions.get(commit.rev_id)
        if not revision or revision["fields"]["repositoryPHID"] == repo.phid:
            continue

        # When uplifting, ensure the `Differential Revision` line is properly
        # moved, and that `rev-id` is updated to not point at the original revision.
        commit.body, commit.rev_id = move_drev_to_original(
            commit.body,
            commit.rev_id,
        )


def update_revision_description(
    transactions: List[dict], commit: Commit, revision: dict
):
    # Appends differential.revision.edit transaction(s) to `transactions` if
    # updating the commit title and/or summary is required.

    if commit.title != revision["fields"]["title"]:
        transactions.append({"type": "title", "value": commit.title})

    # The Phabricator API will refuse the new summary value if we include the
    # "Differential Revision:" keyword in the summary body.
    local_body = strip_differential_revision(commit.body).strip()
    remote_body = strip_differential_revision(revision["fields"]["summary"]).strip()
    if local_body != remote_body:
        transactions.append({"type": "summary", "value": local_body})


def update_revision_bug_id(transactions: List[dict], commit: Commit, revision: dict):
    # Appends differential.revision.edit transaction(s) to `transactions` if
    # updating the commit bug-id is required.
    if commit.bug_id and commit.bug_id != revision["fields"]["bugzilla.bug-id"]:
        transactions.append({"type": "bugzilla.bug-id", "value": commit.bug_id})


def local_uplift_if_possible(
    args: argparse.Namespace, repo: Repository, commits: List[Commit]
) -> bool:
    """If possible, rebase local repository commits onto the target uplift train.

    Uplifts will be performed if the `--no-rebase` argument is not
    passed, if the local repository has a corresponding unified head
    for the uplift train, and if the revset being submitted is not
    already descendant from the target unified head.

    Returns a `bool` indicating if the `commits` should avoid making local
    repository changes to reflect new Phabricator revisions. If `True`,
    local commits may be amended to reflect their state on Phabricator,
    for example to add the `Differential Revision` line. If `False`, the
    local repository commits should not be updated by `moz-phab`.
    """
    if args.no_rebase:
        # If args tell us not to do a rebase, do not make any local changes and
        # return without rebasing. This is the same as submitting an uplift where
        # the original patch is sent to Phabricator without any modifications.
        # In this case we want to avoid local amendments to commits.
        return True

    # Try and find a local repo identifier (hg bookmark, git remote branch) to rebase
    # our revset onto.
    unified_head = repo.get_repo_head_branch()

    if not unified_head:
        # If we didn't find a unified head, we intend to submit an uplift without
        # modifying the local repo state via a rebase.
        logger.warning(
            f"Couldn't find a head for {args.train} in version control, "
            "submitting without rebase."
        )
        return True

    if not repo.is_descendant(unified_head):
        # If we found a head to rebase onto and the commit isn't already a descendant
        # of our target identifier, uplift the commits onto the unified head.
        with wait_message(f"Rebasing commits onto {unified_head}"):
            commits = repo.uplift_commits(unified_head, commits)

    return False


def _submit(repo: Repository, args: argparse.Namespace) -> List[Commit]:
    telemetry().submission.preparation_time.start()
    with wait_message("Checking connection to Phabricator."):
        # Check if raw Conduit API can be used
        if not conduit.check():
            raise Error("Failed to use Conduit API")

        # Check if local and remote VCS matches
        repo.check_vcs()

    repo.before_submit()

    # Find and preview commits to submits.
    with wait_message("Looking for commits.."):
        commits = repo.commit_stack(single=args.single)
    if not commits:
        raise Error("Failed to find any commits to submit")
    n_commits = len(commits)
    if n_commits > 100:
        raise Error(
            f"Unable to create a stack with {n_commits} unpublished commits.\n\n"
            "This is usually the result of a failure to detect the correct "
            "remote repository.\nTry again with the `--upstream <upstream>` "
            "switch to specify the correct remote repository,\n"
            "or set the `git.remote` moz-phab config option to specify a remote."
        )

    if args.command == "uplift":
        # Perform uplift logic during submission.
        avoid_local_changes = local_uplift_if_possible(args, repo, commits)
    else:
        avoid_local_changes = False

    with wait_message("Loading commits.."):
        # Pre-process to load metadata.
        morph_blocking_reviewers(commits)
        augment_commits_from_body(commits)
        update_commits_from_args(commits, args)
        if args.command == "uplift":
            update_commits_for_uplift(commits, repo)
        update_commit_title_previews(commits)

    # Display a one-line summary of commit and WIP count.
    commit_count = len(commits)
    wip_commit_count = sum(1 for commit in commits if commit.wip)

    if wip_commit_count == commit_count:
        status = "as Work In Progress"
    elif wip_commit_count:
        status = f"{wip_commit_count} as Work In Progress"
    else:
        status = "for review"

    logger.warning(
        f"Submitting {commit_count} commit{'s'[: commit_count ^ 1]} {status}"
    )

    with wait_message("Checking commits..."):
        warnings, errors = validate_commit_stack(commits, args)
    if errors:
        log_commit_stack_with_messages(commits, errors, "- ", "  ", logging.ERROR)
        raise Error("Unable to submit commits")
    log_commit_stack_with_messages(commits, warnings, "!! ", "   ")

    try:
        repo.check_commits_for_submit(commits)
    except Error as e:
        if not args.force:
            raise Error("Unable to submit commits:\n\n%s" % e)
        logger.error("Ignoring issues found with commits:\n\n%s", e)

    if not any(commit.submit for commit in commits):
        logger.warning("No changes to submit.")
        return []

    # Show a warning if there are untracked files.
    if config.warn_untracked:
        untracked = repo.untracked()
        if untracked:
            logger.warning(
                "Warning: found %s untracked file%s (will not be submitted):",
                len(untracked),
                "" if len(untracked) == 1 else "s",
            )
            if len(untracked) <= 5:
                for filename in untracked:
                    logger.info("  %s", filename)

    # Show a warning if -m is used and there are new commits.
    if args.message and any(commit for commit in commits if not commit.rev_id):
        logger.warning(
            "Warning: --message works with updates only, and will not\n"
            "result in a comment on new revisions."
        )

    telemetry().submission.preparation_time.stop()
    telemetry().submission.commits_count.add(len(commits))

    # Confirmation prompt.
    if args.yes:
        pass
    elif config.auto_submit and not args.interactive:
        logger.info(
            "Automatically submitting (as per submit.auto_submit in %s)",
            config.filename,
        )
    else:
        res = prompt(f"Submit to {repo.phab_url}", ["Yes", "No", "Always"])
        if res == "No":
            return []
        if res == "Always":
            config.auto_submit = True
            config.write()

    # Process.
    telemetry().submission.process_time.start()

    previous_commit = None
    for commit in commits:
        if not commit.submit:
            previous_commit = commit
            continue

        # Only revisions being updated have an ID. Newly created ones don't.
        is_update = bool(commit.rev_id)

        # Let the user know something's happening.
        if is_update:
            logger.info("\nUpdating revision D%s:", commit.rev_id)
        else:
            logger.info("\nCreating new revision:")

        logger.info("%s %s", commit.name, commit.revision_title())

        # Create a diff if needed
        with wait_message("Creating local diff..."):
            diff = repo.get_diff(commit)

        if diff:
            telemetry().submission.files_count.add(len(diff.changes))
            with wait_message("Uploading binary file(s)..."):
                conduit.upload_files_from_diff(diff)

            with wait_message("Submitting the diff..."):
                result = conduit.submit_diff(diff, commit)
                diff.phid = result["phid"]
                diff.id = result["diffid"]

        if is_update:
            with wait_message("Updating revision..."):
                rev = conduit.update_revision(
                    commit,
                    diff_phid=diff.phid,
                    comment=args.message,
                )
        else:
            with wait_message("Creating a new revision..."):
                rev = conduit.create_revision(
                    commit,
                    diff.phid,
                    # Set the parent revision if one is available.
                    parent_rev_phid=(
                        previous_commit.rev_phid if previous_commit else None
                    ),
                )

        # Set revision ID and PHID from the Conduit API response.
        commit.rev_id = rev["object"]["id"]
        commit.rev_phid = rev["object"]["phid"]

        revision_url = "%s/D%s" % (repo.phab_url, commit.rev_id)

        # Append/replace div rev url to/in commit description.
        body = amend_revision_url(commit.body, revision_url)

        # Amend the commit if required.
        # As commit rewriting can be expensive we avoid it in some circumstances, such
        # as pre-pending "WIP: " to commits submitted as WIP to Phabricator.
        if commit.title_preview != commit.title or body != commit.body:
            commit.title = commit.title_preview
            commit.body = body

            if not avoid_local_changes:
                with wait_message("Updating commit.."):
                    repo.amend_commit(commit, commits)

        # Diff property has to be set after potential SHA1 change.
        if diff:
            with wait_message("Setting diff metadata..."):
                message = commit.build_arc_commit_message()
                conduit.set_diff_property(diff.id, commit, message)

        previous_commit = commit

    # Cleanup (eg. strip nodes) and refresh to ensure the stack is right for the
    # final showing.
    with wait_message("Cleaning up.."):
        if args.command != "uplift":
            repo.finalize(commits)
        repo.after_submit()
        repo.cleanup()
        repo.refresh_commit_stack(commits)

    logger.warning("\nCompleted")
    show_commit_stack(commits)
    telemetry().submission.process_time.stop()

    # Indicate that commits were submitted.
    return commits


def submit(repo: Repository, args: argparse.Namespace) -> List[Commit]:
    try:
        return _submit(repo, args)
    except Exception as e:
        # Check if the `fallback` flag was set and print a warning message.
        if hasattr(args, "fallback") and args.fallback:
            logger.warning(
                "You didn't specify a valid command, so we ran `submit` for you, "
                "and it failed."
            )

        raise e


def add_parser(parser):
    submit_parser = parser.add_parser(
        "submit",
        aliases=["upload"],
        help="Submit commit(s) to Phabricator.",
        description=(
            "MozPhab will change the working directory and amend the commits during "
            "the submission process."
        ),
    )
    add_submit_arguments(submit_parser)
    submit_parser.set_defaults(func=submit, needs_repo=True)


def add_submit_arguments(parser):
    """Add `moz-phab submit` command line arguments to a parser."""
    parser.add_argument(
        "--path", "-p", help="Set path to repository (default: detected)."
    )
    parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Submit without confirmation (default: %s)." % config.auto_submit,
    )
    parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        help="Submit with confirmation (default: %s)." % (not config.auto_submit),
    )
    parser.add_argument(
        "--message",
        "-m",
        help="Provide a custom update message (default: none).",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Override sanity checks and force submission; a tool of last resort.",
    )
    parser.add_argument(
        "--force-delete",
        action="store_true",
        help="Mercurial only: Ignore error caused by a DAG branch point without "
        "evolve installed.",
    )
    parser.add_argument(
        "--bug",
        "-b",
        help="Set Bug ID for all commits (default: from commit).",
        type=str,
    )
    parser.add_argument(
        "--no-bug",
        action="store_true",
        help="Continue if a bug number is not provided.",
    )
    parser.add_argument(
        "--reviewer",
        "--reviewers",
        "-r",
        action="append",
        help="Set review(s) for all commits (default: from commit).",
    )
    parser.add_argument(
        "--blocker",
        "--blockers",
        "-R",
        action="append",
        help="Set blocking review(s) for all commits (default: from commit).",
    )
    wip_group = parser.add_mutually_exclusive_group()
    wip_group.add_argument(
        "--wip",
        "--plan-changes",
        action="store_true",
        help="Create or update a revision without requesting a code review.",
    )
    wip_group.add_argument(
        "--no-wip",
        action="store_true",
        help=(
            "Force revisions to not be marked as work-in-progress. The status of "
            "commits without reviewers can be overridden for example."
        ),
    )
    parser.add_argument(
        "--less-context",
        "--lesscontext",
        action="store_true",
        dest="lesscontext",
        help=(
            "Normally, files are diffed with full context: the entire file "
            "is sent to Phabricator so reviewers can 'show more' and see "
            "it. If you are making changes to very large files with tens of "
            "thousands of lines, this may not work well. With this flag, a "
            "revision will be created that has only a few lines of context."
        ),
    )
    parser.add_argument(
        "--no-stack",
        action="store_true",
        help="Submit multiple commits, but do not mark them as dependent.",
    )
    parser.add_argument(
        "--upstream",
        "--remote",
        "-u",
        action="append",
        help=(
            "Git only: Set remote to detect the starting commit. Overrides "
            "`git.remote` in config."
        ),
    )
    parser.add_argument(
        "--force-vcs",
        action="store_true",
        help="EXPERIMENTAL: Override VCS compatibility check.",
    )
    parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    parser.add_argument(
        "--single",
        "-s",
        action="store_true",
        help="Submit a single commit.",
    )
    parser.add_argument(
        "start_rev",
        nargs="?",
        default=environment.DEFAULT_START_REV,
        help="Start revision of range to submit (default: detected).",
    )
    parser.add_argument(
        "end_rev",
        nargs="?",
        default=environment.DEFAULT_END_REV,
        help="End revision of range to submit (default: current commit).",
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest import mock

from mozphab.commands.submit import (
    local_uplift_if_possible,
    update_commits_for_uplift,
)
from mozphab.commands.uplift import (
    build_assessment_linking_url,
)
from mozphab.commits import Commit
from mozphab.helpers import ORIGINAL_DIFF_REV_RE


class Repo:
    def __init__(self, unified_head="beta", is_descendant=True, phid="PHID-beta"):
        self.unified_head = unified_head
        self._is_descendant = is_descendant
        self.uplift_called = False
        self.phid = phid

    def get_repo_head_branch(self):
        return self.unified_head

    def is_descendant(self, *args, **kwargs):
        return self._is_descendant

    def uplift_commits(self, *args, **kwargs):
        self.uplift_called = True


def test_local_uplift_if_possible():
    class Args:
        def __init__(self, no_rebase=False, train="train"):
            self.no_rebase = no_rebase
            self.train = train

    commits = [
        Commit(
            title="A",
            reviewers={"granted": ["john"], "request": []},
            bug_id=None,
            body="",
            rev_id=1,
        ),
    ]

    repo = Repo()

    args = Args(no_rebase=True)
    assert (
        local_uplift_if_possible(args, repo, commits) is True
    ), "Should always do a one-off uplift when `--no-rebase` is set."

    args = Args()
    repo = Repo(unified_head=None)

    assert (
        local_uplift_if_possible(args, repo, commits) is True
    ), "Should avoid do a one-off when no unified head is found."

    repo = Repo(is_descendant=True)
    assert (
        local_uplift_if_possible(args, repo, commits) is False
    ), "Should avoid uplifting commits locally when destination is a descendant."

    # Rebase-uplift case.
    repo = Repo(
        is_descendant=False,
        unified_head="beta",
    )
    args = Args(
        no_rebase=False,
        train="beta",
    )
    assert (
        local_uplift_if_possible(args, repo, commits) is False
    ), "Uplifting commits locally should amend them as well."
    assert (
        repo.uplift_called
    ), "Should call `uplift_commits` when non-descendant unified head found."


def test_update_commits_for_uplift_sets_relman_review():
    commits = [
        Commit(
            title="A",
            reviewers={"granted": ["john"], "request": []},
            bug_id=None,
            body="",
            rev_id=None,
        ),
        Commit(
            title="B",
            reviewers={"granted": ["john"], "request": ["doe"]},
            bug_id=None,
            body="",
            rev_id=None,
        ),
    ]

    update_commits_for_uplift(commits, Repo())

    reviewers = commits[0].reviewers

    assert not reviewers[
        "request"
    ], "Uplifted patch should have no requested reviewers initially."
    assert not reviewers[
        "granted"
    ], "Uplifted patch should have no granted reviewers initially."

    reviewers = commits[1].reviewers

    assert not reviewers[
        "request"
    ], "Uplifted patch should have no requested reviewers initially."
    assert not reviewers[
        "granted"
    ], "Uplifted patch should have no granted reviewers initially."


def test_update_commits_for_uplift_sets_original_revision():
    commits = [
        # Check initial submission.
        Commit(
            title="bug 1: firstline r?reviewer",
            reviewers={"granted": ["john"], "request": []},
            bug_id="1",
            body=(
                "bug 1: firstline r?reviewer\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D1\n"
            ),
            rev_id=1,
        ),
        # Check update of existing uplift revision.
        Commit(
            title="bug 1: firstline r?reviewer",
            reviewers={"granted": ["john"], "request": []},
            bug_id="1",
            body=(
                "bug 1: firstline r?reviewer\n"
                "\n"
                "Original Revision: https://phabricator.services.mozila.com/D1\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D2\n"
            ),
            rev_id=2,
        ),
        # Check another initial submission.
        Commit(
            title="bug 3: commit message",
            reviewers={"granted": [], "request": []},
            bug_id="3",
            body=(
                "bug 3: commit message\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D3\n"
            ),
            rev_id=3,
        ),
    ]

    with mock.patch("mozphab.commands.submit.conduit.get_revisions") as m_get_revs:
        m_get_revs.return_value = [
            {"id": 1, "fields": {"repositoryPHID": "PHID-mc"}},
            {"id": 2, "fields": {"repositoryPHID": "PHID-beta"}},
            {"id": 3, "fields": {"repositoryPHID": "PHID-mc"}},
        ]
        update_commits_for_uplift(commits, Repo())

    # Initial submission.
    body = commits[0].body
    rev_id = commits[0].rev_id

    assert "Differential Revision:" not in body
    assert "Original Revision:" in body
    assert rev_id is None

    # Update of existing uplift.
    body = commits[1].body
    rev_id = commits[1].rev_id

    assert "Differential Revision:" in body
    assert "Original Revision:" in body
    assert rev_id == 2

    # Another initial submission.
    body = commits[2].body
    rev_id = commits[2].rev_id

    assert "Differential Revision:" not in body
    assert ORIGINAL_DIFF_REV_RE.search(body).group("rev") == "3"
    assert rev_id is None


def test_uplift_beta_commit_to_esr():
    commit = Commit(
        title="bug 2: commit message r=john",
        reviewers={"granted": ["john"], "request": []},
        bug_id="2",
        body=(
            "bug 2: commit message r=john\n"
            "\n"
            "Original Revision: https://phabricator.services.mozila.com/D1\n"
            "\n"
            "Differential Revision: https://phabricator.services.mozila.com/D2\n"
        ),
        rev_id=2,
    )

    with mock.patch("mozphab.commands.submit.conduit.get_revisions") as m_get_revs:
        m_get_revs.return_value = [{"id": 2, "fields": {"repositoryPHID": "PHID-beta"}}]
        update_commits_for_uplift([commit], Repo(phid="PHID-esr"))

    reviewers = commit.reviewers
    body = commit.body
    rev_id = commit.rev_id

    assert not reviewers["request"]
    assert not reviewers["granted"]
    assert "Differential Revision:" not in body
    assert ORIGINAL_DIFF_REV_RE.search(body).group("rev") == "1"
    assert rev_id is None


def test_update_commits_for_uplift_strips_dontbuild():
    commits = [
        Commit(
            title="bug 1: fix something r=reviewer DONTBUILD",
            reviewers={"granted": ["reviewer"], "request": []},
            bug_id="1",
            body=(
                "bug 1: fix something r=reviewer DONTBUILD\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D1\n"
            ),
            rev_id=1,
        ),
        Commit(
            title="bug 2: another fix r=reviewer DONTBUILD (NPOTB)",
            reviewers={"granted": ["reviewer"], "request": []},
            bug_id="2",
            body=(
                "bug 2: another fix r=reviewer DONTBUILD (NPOTB)\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D2\n"
            ),
            rev_id=2,
        ),
    ]

    with mock.patch("mozphab.commands.submit.conduit.get_revisions") as m_get_revs:
        m_get_revs.return_value = [
            {"id": 1, "fields": {"repositoryPHID": "PHID-mc"}},
            {"id": 2, "fields": {"repositoryPHID": "PHID-mc"}},
        ]
        update_commits_for_uplift(commits, Repo())

    assert (
        "DONTBUILD" not in commits[0].title
    ), "DONTBUILD should be stripped from title on uplift."
    assert (
        commits[0].title == "bug 1: fix something r=reviewer"
    ), "Title should have DONTBUILD removed."

    assert (
        "DONTBUILD" not in commits[1].title
    ), "DONTBUILD (NPOTB) should be stripped from title on uplift."
    assert (
        commits[1].title == "bug 2: another fix r=reviewer"
    ), "Title should have DONTBUILD (NPOTB) removed."


def test_build_assessment_linking_url():
    url = build_assessment_linking_url("https://lando.moz.tools", 123)
    assert (
        url == "https://lando.moz.tools/uplift/request/?revisions=123"
    ), "Revision ID only should produce expected URL."

    url = build_assessment_linking_url("https://lando.moz.tools", 123, 5)
    assert (
        url == "https://lando.moz.tools/uplift/request/?revisions=123&assessment_id=5"
    ), "Revision ID and assessment ID should produce expected URL."

    url = build_assessment_linking_url("https://lando.moz.tools/", 456)
    assert (
        url == "https://lando.moz.tools/uplift/request/?revisions=456"
    ), "Trailing slash in Lando URL should produce valid URL."

    url = build_assessment_linking_url("https://lando.moz.tools/", 456, 10)
    assert (
        url == "https://lando.moz.tools/uplift/request/?revisions=456&assessment_id=10"
    ), "Trailing slash in Lando URL with assessment ID should produce valid URL."



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import configparser
import io
import os
from pathlib import Path
from typing import (
    Any,
    Optional,
)

from mozphab import environment

from .logger import logger


class Config(object):
    def __init__(
        self, should_access_file: bool = True, filename: Optional[Path] = None
    ):
        """Sets default config and overrides it with values from the config file.

        Kwargs:
            should_access_file (bool) - should the file be read/write?
            filename (string) - file to read/write the config - used only for testing.
        """
        self._should_access_file = should_access_file

        self.filename = filename or (environment.HOME_DIR / ".moz-phab-config")

        # Default values.
        defaults = """
            [ui]
            no_ansi = False
            hyperlinks = True

            [vcs]
            safe_mode = False

            [git]
            remote =
            command_path =

            [hg]
            command_path =

            [submit]
            auto_submit = False
            always_blocking = False
            warn_untracked = True

            [patch]
            apply_to = base
            create_bookmark = True
            create_topic = False
            create_branch = True
            always_full_stack = False
            branch_name_template = phab-D{rev_id}
            create_commit = True

            [updater]
            self_last_check = 0
            self_auto_update = True
            get_pre_releases = False

            [error_reporting]
            report_to_sentry = True

            [telemetry]
            enabled = False
            """

        self._config = configparser.ConfigParser()
        self._config.read_file(
            io.StringIO("\n".join([line.strip() for line in defaults.splitlines()]))
        )

        if self._config.has_section("arc"):
            self._config.remove_section("arc")

        if should_access_file:
            self._config.read([self.filename])

        self.no_ansi = self._getboolean("ui", "no_ansi")
        self.hyperlinks = self._getboolean("ui", "hyperlinks")
        self.safe_mode = self._getboolean("vcs", "safe_mode")
        self.auto_submit = self._getboolean("submit", "auto_submit")
        self.always_blocking = self._getboolean("submit", "always_blocking")
        self.warn_untracked = self._getboolean("submit", "warn_untracked")
        self.apply_patch_to = self._config.get("patch", "apply_to")
        self.create_bookmark = self._getboolean("patch", "create_bookmark")
        self.create_topic = self._getboolean("patch", "create_topic")
        self.create_branch = self._getboolean("patch", "create_branch")
        self.always_full_stack = self._getboolean("patch", "always_full_stack")
        self.branch_name_template = self._config.get("patch", "branch_name_template")
        self.create_commit = self._getboolean("patch", "create_commit")
        self.self_last_check = self._getint("updater", "self_last_check")
        self.self_auto_update = self._getboolean("updater", "self_auto_update")
        self.get_pre_releases = self._getboolean("updater", "get_pre_releases")
        git_remote = self._config.get("git", "remote")
        self.git_remote = git_remote.replace(" ", "").split(",") if git_remote else []
        self.report_to_sentry = self._getboolean("error_reporting", "report_to_sentry")
        self.telemetry_enabled = self._getboolean("telemetry", "enabled")
        self._git_command = self._config.get("git", "command_path")
        self.git_command = (
            [self._git_command] if self._git_command else environment.GIT_COMMAND
        )
        self._hg_command = self._config.get("hg", "command_path")
        self.hg_command = (
            [self._hg_command] if self._hg_command else environment.HG_COMMAND
        )

        if should_access_file and not os.path.exists(self.filename):
            self.write()

    def _set(self, section: str, option: str, value: Any):
        if not self._config.has_section(section):
            self._config.add_section(section)
        self._config.set(section, option, str(value))

    def _getboolean(self, section: str, option: str) -> bool:
        try:
            return self._config.getboolean(section, option)
        except ValueError as e:
            raise ValueError(
                f"could not convert {section}.{option} to a boolean: {str(e)}"
            )

    def _getint(self, section: str, option: str) -> int:
        try:
            return self._config.getint(section, option)
        except ValueError as e:
            raise ValueError(
                f"could not convert {section}.{option} to an integer: {str(e)}"
            )

    def write(self):
        if not self._should_access_file:
            return

        if os.path.exists(self.filename):
            logger.debug("updating %s", self.filename)
            self._set("submit", "auto_submit", self.auto_submit)
            self._set("patch", "always_full_stack", self.always_full_stack)
            self._set("updater", "self_last_check", self.self_last_check)
            self._set("updater", "self_auto_update", self.self_auto_update)
            self._set("updater", "get_pre_releases", self.get_pre_releases)
            self._set("telemetry", "enabled", self.telemetry_enabled)

        else:
            logger.debug("creating %s", self.filename)
            self._set("ui", "no_ansi", self.no_ansi)
            self._set("ui", "hyperlinks", self.hyperlinks)
            self._set("vcs", "safe_mode", self.safe_mode)
            self._set("git", "remote", ", ".join(self.git_remote))
            self._set("git", "command_path", ", ".join(self._git_command))
            self._set("hg", "command_path", ", ".join(self.hg_command))
            self._set("submit", "auto_submit", self.auto_submit)
            self._set("submit", "always_blocking", self.always_blocking)
            self._set("submit", "warn_untracked", self.warn_untracked)
            self._set("patch", "apply_to", self.apply_patch_to)
            self._set("patch", "create_bookmark", self.create_bookmark)
            self._set("patch", "create_topic", self.create_topic)
            self._set("patch", "create_branch", self.create_branch)
            self._set("patch", "always_full_stack", self.always_full_stack)
            self._set("patch", "create_commit", self.create_commit)
            self._set("telemetry", "enabled", self.telemetry_enabled)

        with self.filename.open("w", encoding="utf-8") as f:
            self._config.write(f)


should_access_file = True
if os.environ.get("MOZPHAB_NO_USER_CONFIG", False):
    should_access_file = False

config = Config(should_access_file=should_access_file)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Generate change log to be pasted in the Mozilla wiki and Discourse:
    - https://wiki.mozilla.org/MozPhab/Changelog
    - https://discourse.mozilla.org/c/firefox-tooling-announcements
"""

import argparse
import json
import re
import subprocess
import time
import urllib.request
from pathlib import Path
from typing import List

# If we are `moz-phab/dev/release_announcement.py`, then `moz-phab` is the following.
MOZPHAB_PATH_DEFAULT = Path(__file__).resolve().parent.parent


def fetch_bugzilla_info(bug: str) -> dict:
    """Get bug info from the Bugzilla API
    Args:
        bug (str): A bug ID to query.
    Returns:
        dict: Dictionary containing bug info of the given bug.
    """
    url = f"https://bugzilla.mozilla.org/rest/bug/{bug}"
    output = urllib.request.urlopen(urllib.request.Request(url), timeout=30).read()
    response_json = json.loads(output.decode("utf-8"))

    # TODO -- is there a better way to query individual bugs, or a list of bugs?
    return response_json["bugs"][0]


def get_bug_ids(
    last_version: str,
    current_version: str,
    mozphab_path: Path,
) -> List[str]:
    """Fetch commits between `last_version` and `current_version` and return Bug IDs."""
    output = subprocess.check_output(
        ["git", "log", "--oneline", f"{last_version}..{current_version}"],
        cwd=mozphab_path,
        encoding="utf-8",
    )

    bug_re = re.compile(r"^.*[Bug|bug] (\d+).*$", flags=re.MULTILINE)

    bug_ids = []
    for line in output.split("\n"):
        bug = bug_re.match(line)
        if bug:
            bug_ids.append(bug.groups()[0])

    bug_ids = list(set(bug_ids))
    bug_ids.sort()
    return bug_ids


def discourse_formatted_text(current_version: str, bug_titles: dict) -> str:
    """Return text formatted for Discourse."""
    out = [f"Bugs resolved in Moz-Phab {current_version}:"]
    for bug, bug_title in bug_titles.items():
        out.append(f"- [bug {bug}](https://bugzilla.mozilla.org/{bug}) {bug_title}")
    out.append("")
    out.append("Discuss these changes in #engineering-workflow on Slack or #Conduit Matrix.")

    return "\n".join(out)


def wiki_formatted_text(current_version: str, bug_titles: dict) -> str:
    """Return text formatted for Mozilla Wiki."""
    out = [f"=== {current_version} ==="]
    for bug, bug_title in bug_titles.items():
        out.append(f"* {{{{bug|{bug}}}}} {bug_title}")

    out.append("")

    return "\n".join(out)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "current_version",
        help="Current version of moz-phab to generate announcements for.",
    )
    parser.add_argument("last_version", help="Previously released moz-phab version.")
    parser.add_argument(
        "--mozphab-path",
        help="Path to moz-phab directory.",
        default=MOZPHAB_PATH_DEFAULT,
    )

    args = parser.parse_args()

    bug_ids = get_bug_ids(args.last_version, args.current_version, args.mozphab_path)
    bug_titles = {}

    for bug in bug_ids:
        time.sleep(0.1)
        bug_titles[bug] = fetch_bugzilla_info(bug)["summary"]

    print()
    print(f"{'*' * 32} >8 copy and paste to wiki {'*' * 32}")
    print(wiki_formatted_text(args.current_version, bug_titles))

    print(f"{'*' * 32} >8 copy and paste to discourse {'*' * 32}")
    print(discourse_formatted_text(args.current_version, bug_titles))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab.environment import (
    IS_WINDOWS,
    MOZPHAB_VERSION,
)
from mozphab.logger import logger
from mozphab.spinner import wait_message
from mozphab.updater import (
    check_for_updates,
    log_windows_update_message,
    self_upgrade,
)


def self_update(_):
    """`self-update` command, updates this package"""
    new_version = check_for_updates(force_check=True)

    if not new_version:
        logger.info(
            f"You are running the latest version of `moz-phab`, {MOZPHAB_VERSION}."
        )
        return

    if IS_WINDOWS:
        log_windows_update_message()
        return

    with wait_message(f"Upgrading to version {new_version}"):
        self_upgrade()

    logger.info(f"Upgraded to version {new_version}")


def add_parser(parser):
    update_parser = parser.add_parser("self-update", help="Update review script.")
    update_parser.set_defaults(func=self_update, needs_repo=False)
